#  Copyright (c) 2019 Toyota Research Institute
"""
This module implements agent-tools for meta-analysis of
other agents
"""


import numpy as np
from taburu.table import ParameterTable
import pandas as pd


REGRESSOR_PARAMS = [
    {
        "@class": ["sklearn.linear_model.LinearRegression"],
        "fit_intercept": [True, False],
        "normalize": [True, False],
    },
    {
        "@class": ["sklearn.ensemble.RandomForestRegressor"],
        "n_estimators": [100],
        "max_features": list(np.arange(0.05, 1.01, 0.05)),
        "min_samples_split": list(range(2, 21)),
        "min_samples_leaf": list(range(1, 21)),
        "bootstrap": [True, False],
    },
    {
        "@class": ["sklearn.neural_network.MLPRegressor"],
        "hidden_layer_sizes": [
            # I think there's a better way to support this, but need to think
            (80, 50),
            (84, 55),
            (87, 60),
        ],
        "activation": ["identity", "logistic", "tanh", "relu"],
        "learning_rate": ["constant", "invscaling", "adaptive"],
    },
]


AGENT_PARAMS = [
    {
        "@class": ["camd.agent.stability.QBCStabilityAgent"],
        "n_query": [4, 6, 8],
        "n_members": list(range(2, 5)),
        "hull_distance": list(np.arange(0.05, 0.21, 0.05)),
        "training_fraction": [0.4, 0.5, 0.6],
        "model": REGRESSOR_PARAMS,
    },
    {
        "@class": ["camd.agent.agents.AgentStabilityML5"],
        "n_query": [4, 6, 8],
        "hull_distance": [0.05, 0.1, 0.15, 0.2],
        "exploit_fraction": [0.4, 0.5, 0.6],
        "model": REGRESSOR_PARAMS,
    },
]


def convert_parameter_table_to_dataframe(parameter_table, fillna=np.nan):
    """
    Converts parameter table in its current state to dataframe

    Args:
        parameter_table (ParameterTable): parameter table to
            convert to array
        fillna (int): a fill value for any remaining entries
            in the table

    Returns:
        (DataFrame): dataframe corresponding to parameter table
            data

    """
    df = pd.DataFrame(parameter_table, dtype="int64")
    df.index = ["-".join([str(i) for i in row]) for row in np.array(df)]
    df["agent"] = [
        parameter_table.hydrate_index(i, construct_object=True)
        for i in range(len(parameter_table))
    ]
    df.fillna(fillna)
    return df


if __name__ == "__main__":
    first = ParameterTable(AGENT_PARAMS)
    first.hydrate_index(3)
    first.hydrate_index(3, construct_object=True)



# Copyright Toyota Research Institute 2019
"""
Contains generic agents which should not be constrained
to a particular mode of materials discovery or associated
decision-making logic
"""
import numpy as np
import GPy

from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import KFold, cross_val_score
from sklearn.pipeline import Pipeline
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, ConstantKernel
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor
from sklearn.neural_network import MLPRegressor

from camd.agent.base import HypothesisAgent


class GenericGPUCB(HypothesisAgent):
    """
    Generic Gaussian Process (GP) and upper confidence bound (UCB)
    based agent that tries to maximize a target. This provide a "naive" batch mode BO.
    We strongly recommend using GPBatchUCB instead.
    """

    def __init__(
        self, candidate_data=None, seed_data=None, n_query=None, alpha=1.0, kernel=None
    ):
        """
        Args:
            candidate_data (pandas.DataFrame): data about the candidates to search over. Must have a "target" column,
                    and at least one additional column that can be used as descriptors.
            seed_data (pandas.DataFrame):  data which to fit the Agent to.
            n_query (int): number of queries in allowed. Defaults to 1.
            alpha (float): mixing parameter for uncertainties in UCB. Defaults to 1.0.
            kernel (scikit-learn kernel): Kernel object for the GP. Defaults to ConstantKernel(1.0)*RBF(1.0).
                    See scikit-learn.gaussian_process.kernels for details.
        """
        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.n_query = n_query if n_query else 1
        self.cv_score = np.inf
        self.alpha = alpha
        self.kernel = kernel if kernel else ConstantKernel(1.0) * RBF(1.0)
        super(GenericGPUCB).__init__()

    def get_hypotheses(self, candidate_data, seed_data=None):
        """
        Get hypotheses method for GenericGPUCB agent

        Args:
            candidate_data (pandas.DataFrame): dataframe of candidates
            seed_data (pandas.DataFrame): dataframe of prior data on
                which to fit GPUCB

        Returns:
            (pandas.DataFrame): top candidates from the GPUCB algorithm

        """

        self.candidate_data = candidate_data.drop(columns=["target"], axis=1)
        self.seed_data = seed_data
        X_seed = seed_data.drop(columns=["target"], axis=1, errors="ignore")
        y_seed = seed_data["target"]
        steps = [
            ("scaler", StandardScaler()),
            (
                "GP",
                GaussianProcessRegressor(
                    kernel=self.kernel, normalize_y=True, n_restarts_optimizer=25
                ),
            ),
        ]
        self.pipeline = Pipeline(steps)
        self.cv_score = np.mean(
            -1
            * cross_val_score(self.pipeline, X_seed, y_seed, cv=KFold(3, shuffle=True))
        )
        self.pipeline.fit(X_seed, y_seed)
        t_pred, unc = self.pipeline.predict(self.candidate_data, return_std=True)
        t_pred += unc * self.alpha
        selected = np.argsort(-1.0 * t_pred)[: self.n_query]
        return candidate_data.iloc[selected]


class GPBatchUCB(HypothesisAgent):
    """
    Generic Gaussian Process (GP) and upper confidence bound (UCB)
    based agent that tries to maximize a target that can
    be used in a batch-mode Bayesian Optimization setting.
    This class implements a version of the Batch-UCB algorithm described by
    Desautels et al. Journal of Machine Learning Research 15 (2014) 4053-4103
    which shares the same basis with Srinivas et al. arXiv preprint arXiv:0912.3995 (2009).
    """

    def __init__(
        self,
        candidate_data=None,
        seed_data=None,
        n_query=None,
        mode="batch",
        alpha=1.0,
        kernel=None,
        **kwargs
    ):
        """
        Args:
            candidate_data (pandas.DataFrame): data about the candidates to search over. Must have a "target" column,
                    and at least one additional column that can be used as descriptors.
            seed_data (pandas.DataFrame):  data which to fit the Agent to.
            n_query (int): number of queries in allowed. Defaults to 1.
            mode (str): "batch" or "naive"; corresponding to original BUCB algorithm (where batches
                composed iteratively) and a naive batch algorithm (where batch is composed in one-shot
                based on ranking). "naive" is mostly for benchmarking, but is also faster so might be
                useful large, limiting dataset sizes.
            alpha (float or str): mixing parameter for uncertainties in UCB. If a float is given, agent will
                use the same constant alpha throughout the campaign. Defaults to 1.0. Setting this as 'auto' will
                use the Theorem 1 from Srivanasan et al. to determine the alpha during batch composition.
                'auto' has two parameters, 'delta' and 'premultip', which default to 0.1 and 0.05 respectively,
                but can be modified by passing as kwargs to the agent. if mode is "naive" this can only be a flaot.
            kernel (GPy kernel): Kernel object for the GP. Defaults to RBF.
        """
        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.n_query = n_query if n_query else 1
        self.mode = mode
        self.alpha = alpha
        self.kernel = kernel
        self.kwargs = kwargs
        self.cv_score = np.inf

        super(GPBatchUCB).__init__()

    def get_hypotheses(self, candidate_data, seed_data=None):
        """
        Methods for getting hypotheses according to the GPBatchUCB algorithm

        Args:
            candidate_data (pandas.DataFrame): candidate data
            seed_data (pandas.DataFrame): seed data

        Returns:
            (pandas.DataFrame): selected hypotheses

        """
        self.candidate_data = candidate_data.drop(
            columns=["target"], axis=1, errors="ignore"
        )

        if seed_data is not None:
            self.seed_data = seed_data
        else:
            raise ValueError(
                "GPBatchUCB Agent requires a finite seed as input. "
                "If you are using this as part of a Campaign, consider "
                "the create_seed option."
            )

        fb_start = max(len(self.seed_data), 1)

        if self.kernel is None:
            self.kernel = GPy.kern.RBF(input_dim=self.candidate_data.shape[1])

        X_seed = self.seed_data.drop(columns=["target"], axis=1)
        y_seed = self.seed_data["target"].to_numpy().reshape(-1, 1)

        y_m, y_std = np.mean(y_seed), np.std(y_seed)
        y_seed = (y_seed - y_m) / y_std

        scaler = StandardScaler()
        scaler.fit(X_seed)

        r_seed, r_y, r_candidates = X_seed, y_seed, self.candidate_data

        if self.mode == "batch":
            batch = []
            for i in range(min(self.n_query, len(self.candidate_data))):
                x = scaler.transform(r_seed).astype(np.float64)
                y = r_y.astype(np.float64).reshape(-1, 1)

                m = GPy.models.GPRegression(
                    x,
                    y,
                    kernel=self.kernel,
                    noise_var=self.kwargs.get("noise_var", 1.0),
                )
                m.optimize(
                    optimizer=self.kwargs.get("optimizer", "bfgs"),
                    max_iters=self.kwargs.get("max_iters", 1000),
                )
                self.kernel = m.kern

                y_pred, var = m.predict(
                    scaler.transform(r_candidates.to_numpy().astype(np.float64))
                )
                t_pred = y_pred * y_std + y_m
                unc = np.sqrt(var) * y_std + y_m

                if self.alpha == "auto":
                    _t = i + fb_start
                    alpha = self.kwargs.get("premultip", 0.05) * np.sqrt(
                        2
                        * np.log(
                            len(self.candidate_data)
                            * _t**2
                            * np.pi**2
                            / 6
                            / self.kwargs.get("delta", 0.1)
                        )
                    )
                    print("- alpha.{}: ".format(i), alpha)
                else:
                    alpha = self.alpha

                t_pred += unc * alpha
                s = np.argmax(t_pred)

                name = r_candidates.index.tolist()[s]
                batch.append(name)
                r_seed = r_seed.append(r_candidates.loc[name])
                r_y = np.append(r_y, np.array([y_pred[s]]).reshape(1, 1), axis=0)
                r_candidates = r_candidates.drop(name)

        elif self.mode == "naive":
            x = scaler.transform(r_seed).astype(np.float64)
            y = r_y.astype(np.float64).reshape(-1, 1)
            m = GPy.models.GPRegression(
                x, y, kernel=self.kernel, noise_var=self.kwargs.get("noise_var", 1.0)
            )
            self.kernel = m.kern
            m.optimize(
                optimizer=self.kwargs.get("optimizer", "bfgs"),
                max_iters=self.kwargs.get("max_iters", 1000),
            )
            y_pred, var = m.predict(
                scaler.transform(r_candidates.to_numpy().astype(np.float64))
            )
            t_pred = y_pred * y_std + y_m
            unc = np.sqrt(var) * y_std + y_m
            t_pred += unc * self.alpha
            indices = np.argsort(-1.0 * t_pred.flatten())[: self.n_query]
            batch = r_candidates.index.to_numpy()[indices].tolist()
        else:
            raise NotImplementedError("Unknown mode for GPBatchUCB Agent")

        return self.candidate_data.loc[batch]


class RegressorAgent(HypothesisAgent):
    """
    Linear regression based agent that tries to maximize a target.
    Best for simple checks and benchmarks.
    """

    def __init__(
        self,
        model,
        features=None,
        target="target",
        candidate_data=None,
        seed_data=None,
        n_query: int = None,
    ):

        """
        Args:
            model (sklearn.RegressorMixin): some regressor with "fit" method
            candidate_data (pandas.DataFrame): data about the candidates to search over. Must have a "target" column,
                    and at least one additional column that can be used as descriptors.
            seed_data (pandas.DataFrame):  data which to fit the Agent to.
            n_query (int): number of queries in allowed. Defaults to 1.
        """
        self.model = model
        self.features = features
        self.target = target
        self.candidate_data = candidate_data
        self.seed_data = seed_data
        self.n_query = n_query if n_query else 1
        super(RegressorAgent).__init__()

    def get_hypotheses(self, candidate_data, seed_data=None):
        """
        Methods for getting hypotheses using linear regression

        Args:
            candidate_data (pandas.DataFrame): candidate data
            seed_data (pandas.DataFrame): seed data

        Returns:
            (pandas.DataFrame): selected hypotheses

        """
        # Fit on known data
        self.candidate_data = candidate_data

        if seed_data is not None:
            self.seed_data = seed_data
        else:
            raise ValueError(
                "RegressorAgent requires a finite seed as input. "
                "If you are using this as part of a Campaign, consider "
                "the create_seed option."
            )

        if self.features is not None:
            X_seed = seed_data[self.features]
            X_cand = candidate_data[self.features]
        else:
            X_seed = seed_data.drop(columns=[self.target], axis=1)
            X_cand = candidate_data.drop(
                columns=[self.target], axis=1, errors="ignore"
            )
        y_seed = seed_data[self.target]
        self.model.fit(X_seed, y_seed)
        output = self.model.predict(X_cand)
        sorted_output = np.argsort(output)[::-1]
        selected = sorted_output[: self.n_query]
        return candidate_data.iloc[selected]

    @classmethod
    def from_linear(
        cls,
        features=None,
        target="target",
        candidate_data=None,
        seed_data=None,
        n_query: int = None,
        **kwargs
    ):
        """Preset factory method for a Linear Agent"""
        linear_reg = LinearRegression(**kwargs)
        return cls(
            model=linear_reg,
            features=features,
            target=target,
            candidate_data=candidate_data,
            seed_data=seed_data,
            n_query=n_query,
        )

    @classmethod
    def from_random_forest(
        cls,
        features=None,
        target="target",
        candidate_data=None,
        seed_data=None,
        n_query: int = None,
        **kwargs
    ):
        """Preset factory method for a RandomForestRegressor-based Agent"""
        rf = RandomForestRegressor(**kwargs)
        return cls(
            model=rf,
            features=features,
            target=target,
            candidate_data=candidate_data,
            seed_data=seed_data,
            n_query=n_query,
        )

    @classmethod
    def from_mlp(
        cls,
        features=None,
        target="target",
        candidate_data=None,
        seed_data=None,
        n_query: int = None,
        **kwargs
    ):
        """Preset factory method for an MLP-based Agent"""
        mlp = MLPRegressor(**kwargs)
        return cls(
            model=mlp,
            features=features,
            target=target,
            candidate_data=candidate_data,
            seed_data=seed_data,
            n_query=n_query,
        )



"""
Module for generic utilities for data munging, s3 interaction etc.
"""
import datetime
import re


def get_new_version(current_ver):
    """
    Helper function to increment semantic versioning by date

    Args:
        current_ver (str): string-formatted date as YYYY.MM.DD,
            or YYYY.MM.DD-post{NUMBER} for post-versions released
            on the same day

    Returns:
        (str) new version string

    """
    today = datetime.datetime.today().strftime("%Y.%-m.%-d")

    # Extract current_ver_date
    current_ver_date = re.sub(r"-post\d+", "", current_ver)

    if today == current_ver_date:
        if "post" in current_ver:
            # Increment post by 1
            new_ver = re.sub(r"post(\d+)", lambda exp: "post{}".format(int(exp.group(1)) + 1), current_ver)
        else:
            # Append "-post0" to version
            new_ver = "{}-post0".format(current_ver)
    else:
        # Set new version as today
        new_ver = today

    return new_ver



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.
"""
This module consolidates s3-based dataset loading
for CAMD, mostly in the form of dataframes
"""

import os

import boto3
import botocore
import requests
import pandas as pd
import numpy as np
from pymatgen.core.composition import Composition
from pymatgen.core.structure import Structure
from pymatgen.ext.matproj import MPRester
from monty.os import makedirs_p
from matminer.featurizers.base import MultipleFeaturizer
from matminer.featurizers.composition import (
    ElementProperty,
    Stoichiometry,
    ValenceOrbital,
    IonProperty,
)
from matminer.featurizers.structure import (
    SiteStatsFingerprint,
    StructuralHeterogeneity,
    ChemicalOrdering,
    StructureComposition,
    MaximumPackingEfficiency,
)
from camd import CAMD_CACHE, tqdm


def load_dataframe(dataset_name):
    """
    Helper function to load a dataframe from the utilities below

    Supported datasets include:
        * oqmd_1.2_voronoi_magpie_fingerprints
        * oqmd1.2_exp_based_entries_featurized_v2,

    Args:
        dataset_name (str): dataset name to load

    Returns:
        (DataFrame): dataframe corresponding to specified dataset

    """
    filename = "{}.pickle".format(dataset_name)
    cache_matrio_data(filename)
    return pd.read_pickle(os.path.join(CAMD_CACHE, filename))


def load_default_atf_data():
    """
    Convenience function to load the default test ATF data
    which is a dataframe of OQMD data of binary compounds with
    magpie and voronoi analysis features

    Returns:
        (DataFrame): dataframe of OQMD data and the corresponding
            features

    """
    df = load_dataframe("oqmd_1.2_voronoi_magpie_fingerprints")
    return df[df["N_species"] == 2].sample(frac=0.2)


def filter_dataframe_by_composition(df, composition, formula_column="Composition"):
    """
    Filters dataframe by composition, i. e. finds all
    rows in dataframe where the Composition contains a
    subset of input composition

    Args:
        df (DataFrame): dataframe
        composition (Composition or str): composition
            or formula by which to filter
        formula_column (str): formula column name, defaults
            to "Composition"

    Returns:
        (DataFrame): dataframe where every composition is sampled such
            that its composition is a subset of the input element set

    """
    # Get elements in formula, composition, then filter
    chemsys = set(Composition(composition).keys())
    all_comps = df[formula_column].apply(Composition)
    indices_to_include = [
        ind for ind, comp in all_comps.items() if comp.keys() <= chemsys
    ]
    return df.loc[indices_to_include]


def get_oqmd_data_by_chemsys(chemsys, drop_duplicates=True):
    """
    Utility function for loading a chemsys from cached
    OQMD data, used primarily for campaign simulations

    Args:
        chemsys (str): formula or hyphen separated list
            of elements, e. g. FeO2, Fe-Cl, etc.
        drop_duplicates (bool): whether to drop duplicate
            values from dataframe

    Returns:
        (DataFrame): dataframe corresponding to oqmd
            data from chemsys

    """
    all_data = load_dataframe("oqmd_1.2_voronoi_magpie_fingerprints")
    dataset = filter_dataframe_by_composition(all_data, chemsys.replace("-", ""))
    if drop_duplicates:
        dataset = dataset.drop_duplicates(keep="first")
    return dataset


def get_mp_ele_ref(element):
    """
    Helper function to get the MP compatible elemental references
    to calculate the formation energy
    Args:
        element (str): the element
    Returns:
        the minimum energy of the unary phase of the element from MP
    """
    with MPRester() as mpr:
        energies = [e.energy_per_atom for e in mpr.get_entries_in_chemsys([element])]
    return min(energies)


MP_REFERENCES = {
    "H": -3.39271585,
    "He": -0.00902216,
    "Li": -1.9089228533333333,
    "Be": -3.739412865,
    "B": -6.679390641666667,
    "C": -9.22676982,
    "N": -8.336493965,
    "O": -4.94795546875,
    "F": -1.9114557725,
    "Ne": -0.02582742,
    "Na": -1.312223005,
    "Mg": -1.5968959166666667,
    "Al": -3.74557583,
    "Si": -5.423390655,
    "P": -5.413285861428571,
    "S": -4.136449866875,
    "Cl": -1.8485262325,
    "Ar": -0.06880822,
    "K": -1.110398947,
    "Ca": -1.999462735,
    "Sc": -6.332469105,
    "Ti": -7.895052840000001,
    "V": -9.08235617,
    "Cr": -9.65304747,
    "Mn": -9.161706470344827,
    "Fe": -8.46929704,
    "Co": -7.108317795,
    "Ni": -5.7798218,
    "Cu": -4.09920667,
    "Zn": -1.259460605,
    "Ga": -3.02808992,
    "Ge": -4.61751163,
    "As": -4.65850806,
    "Se": -3.49591147765625,
    "Br": -1.63692654,
    "Kr": -0.05671467,
    "Rb": -0.9805340725,
    "Sr": -1.6894812533333334,
    "Y": -6.466074656666667,
    "Zr": -8.54770063,
    "Nb": -10.10130504,
    "Mo": -10.84563514,
    "Tc": -10.36061991,
    "Ru": -9.27438911,
    "Rh": -7.33850956,
    "Pd": -5.17648694,
    "Ag": -2.8325290566666665,
    "Cd": -0.90620278,
    "In": -2.75168373,
    "Sn": -3.99229498,
    "Sb": -4.128999585,
    "Te": -3.14330093,
    "I": -1.524009065,
    "Xe": -0.03617417,
    "Cs": -0.8954023720689656,
    "Ba": -1.91897083,
    "La": -4.936007105,
    "Ce": -5.933089155,
    "Pr": -4.780899145,
    "Nd": -4.76814321,
    "Pm": -4.7505423225,
    "Sm": -4.717682476666667,
    "Eu": -10.292043475,
    "Gd": -14.07612224,
    "Tb": -4.6343661,
    "Dy": -4.60678684,
    "Ho": -4.58240887,
    "Er": -4.5674248,
    "Tm": -4.475058396666666,
    "Yb": -1.5395952733333333,
    "Lu": -4.52095052,
    "Hf": -9.95718477,
    "Ta": -11.85777728,
    "W": -12.95812647,
    "Re": -12.444527185,
    "Os": -11.22733445,
    "Ir": -8.83843042,
    "Pt": -6.07090771,
    "Au": -3.27388154,
    "Hg": -0.30362902,
    "Tl": -2.36165292,
    "Pb": -3.71258955,
    "Bi": -3.88641638,
    "Ac": -4.1211750075,
    "Th": -7.41385825,
    "Pa": -9.51466466,
    "U": -11.29141001,
    "Np": -12.94777968125,
    "Pu": -14.26782579,
}

QMPY_REFERENCES = {
    "Ac": -4.1060035325,
    "Ag": -2.8217729525,
    "Al": -3.74573946,
    "Ar": -0.00636995,
    "As": -4.651918435,
    "Au": -3.26680174,
    "B": -6.67796758,
    "Ba": -1.92352708,
    "Be": -3.75520865,
    "Bi": -4.038931855,
    "Br": -1.31759562258416,
    "C": -9.2170759925,
    "Ca": -1.977817,
    "Cd": -0.90043514,
    "Ce": -4.7771708225,
    "Cl": -1.47561368438088,
    "Co": -7.089565,
    "Cr": -9.50844998,
    "Cs": -0.85462775,
    "Cu": -3.7159594,
    "Dy": -4.60150328333333,
    "Er": -4.56334055,
    "Eu": -1.8875732,
    "F": -1.45692429086889,
    "Fe": -8.3078978,
    "Ga": -3.031846515,
    "Gd": -4.6550712925,
    "Ge": -4.623692585,
    "H": -3.38063384781582,
    "He": -0.004303435,
    "Hf": -9.955368785,
    "Hg": -0.358963825033731,
    "Ho": -4.57679364666667,
    "I": -1.35196205757168,
    "In": -2.71993876,
    "Ir": -8.8549203,
    "K": -1.096699335,
    "Kr": -0.004058825,
    "La": -4.93543556,
    "Li": -1.89660627,
    "Lu": -4.524181525,
    "Mg": -1.54251595083333,
    "Mn": -9.0269032462069,
    "Mo": -10.8480839,
    "N": -8.11974103465649,
    "Na": -1.19920373914835,
    "Nb": -10.09391206,
    "Nd": -4.762916335,
    "Ne": -0.02931791,
    "Ni": -5.56661952,
    "Np": -12.94027372125,
    "O": -4.52329546412125,
    "Os": -11.22597601,
    "P": -5.15856496104006,
    "Pa": -9.49577589,
    "Pb": -3.70396484,
    "Pd": -5.17671826,
    "Pm": -4.7452352875,
    "Pr": -4.7748066125,
    "Pt": -6.05575959,
    "Pu": -14.29838348,
    "Rb": -0.9630733,
    "Re": -12.422818875,
    "Rh": -7.26940476,
    "Ru": -9.2019888,
    "S": -3.83888286598664,
    "Sb": -4.117563025,
    "Sc": -6.328367185,
    "Se": -3.48117276,
    "Si": -5.424892535,
    "Sm": -4.7147675825,
    "Sn": -3.9140929231488,
    "Sr": -1.6829138,
    "Ta": -11.85252937,
    "Tb": -5.28775675533333,
    "Tc": -10.360747885,
    "Te": -3.14184237666667,
    "Th": -7.41301719,
    "Ti": -7.69805778621374,
    "Tl": -2.359420025,
    "Tm": -4.47502416,
    "U": -11.292348705,
    "V": -8.94097896,
    "W": -12.96020695,
    "Xe": 0.00306349,
    "Y": -6.464420635,
    "Yb": -1.51277545,
    "Zn": -1.2660268,
    "Zr": -8.54717235,
}

QMPY_REFERENCES_HUBBARD = {
    "Co": 2.0736240219357,
    "Cr": 2.79591214925926,
    "Cu": 1.457571831687,
    "Fe": 2.24490453841424,
    "Mn": 2.08652912841877,
    "Ni": 2.56766185643768,
    "Np": 2.77764768949249,
    "Pu": 2.2108747749433,
    "Th": 1.06653674624248,
    "U": 2.57513786752409,
    "V": 2.67812162528461,
}

ELEMENTS = [
    "Ru",
    "Re",
    "Rb",
    "Rh",
    "Be",
    "Ba",
    "Bi",
    "Br",
    "H",
    "P",
    "Os",
    "Ge",
    "Gd",
    "Ga",
    "Pr",
    "Pt",
    "Pu",
    "C",
    "Pb",
    "Pa",
    "Pd",
    "Xe",
    "Pm",
    "Ho",
    "Hf",
    "Hg",
    "He",
    "Mg",
    "K",
    "Mn",
    "O",
    "S",
    "W",
    "Zn",
    "Eu",
    "Zr",
    "Er",
    "Ni",
    "Na",
    "Nb",
    "Nd",
    "Ne",
    "Np",
    "Fe",
    "B",
    "F",
    "Sr",
    "N",
    "Kr",
    "Si",
    "Sn",
    "Sm",
    "V",
    "Sc",
    "Sb",
    "Se",
    "Co",
    "Cl",
    "Ca",
    "Ce",
    "Cd",
    "Tm",
    "Cs",
    "Cr",
    "Cu",
    "La",
    "Li",
    "Tl",
    "Lu",
    "Th",
    "Ti",
    "Te",
    "Tb",
    "Tc",
    "Ta",
    "Yb",
    "Dy",
    "I",
    "U",
    "Y",
    "Ac",
    "Ag",
    "Ir",
    "Al",
    "As",
    "Ar",
    "Au",
    "In",
    "Mo",
]


def cache_download(url, path):
    """
    Quick helper function to cache a generic download from a url
    in the CAMD local data directory

    Args:
        url (str): url for download
        path (str): path for download, is appended to the
            CAMD_CACHE location

    Returns:
        (None)
    """
    # Prep cache path and make necessary dirs
    cache_path = os.path.join(CAMD_CACHE, path)

    # Download and write file
    if not os.path.isfile(cache_path):
        makedirs_p(os.path.split(cache_path)[0])
        r = requests.get(url, stream=True)
        total_size = int(r.headers.get("content-length", 0))
        block_size = 1024  # 1 Kibibyte
        t = tqdm(total=total_size, unit="iB", unit_scale=True)
        with open(cache_path, "wb") as f:
            for data in r.iter_content(block_size):
                t.update(len(data))
                f.write(data)


# Mapping of matrio data hashes to cache paths
MATRIO_DATA_KEYS = {
    "oqmd1.2_exp_based_entries_featurized_v2.pickle": "5e3b0e9bc91e209071f33ce8",
    "oqmd_1.2_voronoi_magpie_fingerprints.pickle": "5e39ce2cd9f13e075b7dfaaf",
    "oqmd_ver3.db": "5e39ce96d9f13e075b7dfab3",
    "oqmd1.2_exp_based_entries_structures.json": "5e45befef23b399192b35242",
}


def get_chemsys(formula_or_structure, seperator="-"):
    """
    Gets a sorted, character-delimited set of elements, e.g.
    Fe-Ni-O or O-Ti

    Args:
        formula_or_structure (str, Structure): formula or structure
            for which to get chemical system
        separator (str): separator for the chemsys elements

    Returns:
        (str): separated

    """
    if isinstance(formula_or_structure, Structure):
        formula = formula_or_structure.composition.reduced_formula
    else:
        formula = formula_or_structure
    elements = [str(el) for el in Composition(formula)]
    return seperator.join(sorted(elements))


def cache_matrio_data(filename):
    """
    Helper function to cache data hosted on data.matr.io

    Args:
        filename (filename): filename to fetch from data.matr.io

    Returns:
        None

    """
    prefix = "https://data.matr.io/3/api/v1/file"
    key = MATRIO_DATA_KEYS[filename]
    if not os.path.isfile(filename):
        cache_download("{}/{}/download".format(prefix, key), filename)


def partition_intercomp(dataframe, n_elements=None):
    """
    Utility function to partition a dataframe into
    data in the interior of the phase diagram

    Args:
        dataframe (DataFrame): dataframe to be partitioned
        n_elements (int): number of elements by which to
            partition the dataframe, defaults to n-1, where n
            is the total number of elements in the dataframe

    Returns:
        (DataFrame): data which contain elements in the
            interior of the phase space according to the
            n_element threshold
        (DataFrame): data which are below the n_element
            threshold
    """
    all_n_elts = [len(Composition(formula)) for formula in dataframe["Composition"]]
    if n_elements is None:
        n_elements = max(all_n_elts) - 1

    mask = np.array(all_n_elts) > n_elements
    return dataframe[mask], dataframe[~mask]


def s3_sync(s3_bucket, s3_prefix, sync_path="."):
    """
    Syncs a given path to an s3 prefix

    Args:
        s3_bucket (str): bucket name
        s3_prefix (str): s3 prefix to sync to
        sync_path (str, Path): path to sync to bucket:prefix

    Returns:
        (None)

    """
    # Get bucket
    s3_resource = boto3.resource("s3")
    bucket = s3_resource.Bucket(s3_bucket)

    # Walk paths and subdirectories, uploading files
    for path, subdirs, files in os.walk(sync_path):
        # Get relative path prefix
        relpath = os.path.relpath(path, sync_path)
        if not relpath.startswith("."):
            prefix = os.path.join(s3_prefix, relpath)
        else:
            prefix = s3_prefix

        for file in files:
            file_key = os.path.join(prefix, file)
            bucket.upload_file(os.path.join(path, file), file_key)


def s3_key_exists(key, bucket):
    """
    Quick utility to determine whether key exists in bucket

    Args:
        key (str): key to check
        bucket (str): bucket to check in

    Returns:
        (bool): whether the key was found in the bucket

    """
    s3 = boto3.resource("s3")
    try:
        s3.Object(bucket, key).load()
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "404":
            # The object does not exist.
            return False
        else:
            # Something else has gone wrong.
            raise e
    else:
        # The object does exist.
        return True


def download_s3_file(key, bucket, output_filename):
    """
    Quick utility to download s3 file

    Args:
        key (str): key to download
        bucket (str): bucket from which to download
        output_filename (str): output filename for object

    Returns:
        (bool): whether the key was found in the bucket

    """
    s3_client = boto3.client("s3")
    s3_client.download_file(bucket, key, output_filename)
    return True


def upload_s3_file(key, bucket, filename):
    """
    Quick utility to upload s3 file

    Args:
        key (str): key to download
        bucket (str): bucket from which to download
        filename (str): output filename for object

    Returns:
        (bool): whether the key was found in the bucket

    """
    s3_client = boto3.client("s3")
    s3_client.upload_file(filename, bucket, key)
    return True


def get_common_prefixes(bucket, prefix):
    """
    Helper function to get common "subfolders" of folders
    in S3

    Args:
        bucket (str): bucket name
        prefix (str): prefix for which to list common prefixes

    Returns:

    """
    if not prefix.endswith("/"):
        prefix += "/"
    client = boto3.client("s3")
    paginator = client.get_paginator("list_objects")
    result = paginator.paginate(Bucket=bucket, Delimiter="/", Prefix=prefix)
    return [
        common_prefix["Prefix"].split("/")[-2]
        for common_prefix in result.search("CommonPrefixes")
        if common_prefix
    ]


def get_default_featurizer():
    """
    Utility function to get CAMD's default featurizer from Ward et al. (2017)

    Returns:
        (Featurizer): default CAMD featurizer

    """
    return MultipleFeaturizer(
        [
            SiteStatsFingerprint.from_preset("CoordinationNumber_ward-prb-2017"),
            StructuralHeterogeneity(),
            ChemicalOrdering(),
            MaximumPackingEfficiency(),
            SiteStatsFingerprint.from_preset("LocalPropertyDifference_ward-prb-2017"),
            StructureComposition(Stoichiometry()),
            StructureComposition(ElementProperty.from_preset("magpie")),
            StructureComposition(ValenceOrbital(props=["frac"])),
            StructureComposition(IonProperty(fast=True)),
        ]
    )



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.
"""
This module provides objects related to the discovery of
new crystal structures using structural domains.
"""

import logging
import os
import pandas as pd
from datetime import datetime
from monty.serialization import dumpfn
from camd import CAMD_TEST_FILES
from camd.agent.base import RandomAgent
from camd.domain import StructureDomain, heuristic_setup
from camd.agent.stability import AgentStabilityAdaBoost
from camd.campaigns.base import Campaign
from camd import CAMD_S3_BUCKET, __version__
from camd.utils.data import load_dataframe, s3_sync, s3_key_exists, \
    upload_s3_file, download_s3_file
from camd.analysis import StabilityAnalyzer
from camd.experiment.dft import OqmdDFTonMC1
from camd.experiment.base import ATFSampler
from sklearn.neural_network import MLPRegressor
from watchtower import CloudWatchLogHandler


class ProtoDFTCampaign(Campaign):
    """
    Subclass of Campaign which implements custom methods
    and factories for constructing prototype-generation
    stability campaigns for materials discovery with DFT
    experiments
    """
    @classmethod
    def from_chemsys(cls, chemsys, prefix="proto-dft-2/runs",
                     n_max_atoms=20, agent=None, analyzer=None,
                     experiment=None, log_file="campaign.log",
                     cloudwatch_group="/camd/worker/dev/"):
        """
        Class factory method for constructing campaign from
        chemsys.

        Args:
            chemsys (str): chemical system for the campaign
            prefix (str): prefix for s3
            n_max_atoms (int): number of maximum atoms
            agent (Agent): agent for stability campaign
            analyzer (Analyzer): analyzer for stability campaign
            experiment (Agent): experiment for stability campaign
            log_file (str): log filename
            cloudwatch_group (str): cloudwatch group to log to

        Returns:
            (ProtoDFTCampaign): Standard proto-dft campaign from
                the chemical system

        """
        logger = logging.Logger("camd")
        logger.setLevel("INFO")
        file_handler = logging.FileHandler(log_file)
        cw_handler = CloudWatchLogHandler(
            log_group=cloudwatch_group,
            stream_name=chemsys
        )
        logger.addHandler(file_handler)
        logger.addHandler(cw_handler)
        logger.addHandler(logging.StreamHandler())

        logger.info("Starting campaign factory from_chemsys {}".format(
            chemsys)
        )
        s3_prefix = "{}/{}".format(prefix, chemsys)

        # Initialize s3
        dumpfn({"started": datetime.now().isoformat(),
                "version": __version__}, "start.json")
        s3_sync(s3_bucket=CAMD_S3_BUCKET, s3_prefix=s3_prefix, sync_path='.')

        # Get structure domain
        # Check cache
        cache_key = "protosearch_cache/v1/{}/{}/candidate_data.pickle".format(chemsys, n_max_atoms)
        # TODO: create test of isfile
        if s3_key_exists(bucket=CAMD_S3_BUCKET, key=cache_key):
            logger.info("Found cached protosearch domain.")
            download_s3_file(cache_key, CAMD_S3_BUCKET, "candidate_data.pickle")
            candidate_data = pd.read_pickle("candidate_data.pickle")
            logger.info("Loaded cached {}.".format(cache_key))
        else:
            logger.info("Generating domain with max {} atoms.".format(n_max_atoms))
            element_list = chemsys.split('-')
            max_coeff, charge_balanced = heuristic_setup(element_list)
            domain = StructureDomain.from_bounds(
                element_list, charge_balanced=charge_balanced,
                n_max_atoms=n_max_atoms, **{'grid': range(1, max_coeff)})
            candidate_data = domain.candidates()
            logger.info("Candidates generated")
            candidate_data.to_pickle("candidate_data.pickle")
            upload_s3_file(cache_key, CAMD_S3_BUCKET, "candidate_data.pickle")
            logger.info("Cached protosearch domain at {}.".format(cache_key))

        # Dump structure/candidate data

        s3_sync(s3_bucket=CAMD_S3_BUCKET, s3_prefix=s3_prefix, sync_path='.')

        # Set up agents and loop parameters
        agent = agent or AgentStabilityAdaBoost(
            model=MLPRegressor(hidden_layer_sizes=(84, 50)),
            n_query=10,
            hull_distance=0.2,
            exploit_fraction=1.0,
            uncertainty=True,
            alpha=0.5,
            diversify=True,
            n_estimators=20
        )
        analyzer = analyzer or StabilityAnalyzer(hull_distance=0.2)
        experiment = experiment or OqmdDFTonMC1(timeout=30000, prefix_append="proto-dft")
        seed_data = load_dataframe("oqmd1.2_exp_based_entries_featurized_v2")

        # Load cached experiments
        logger.info("Loading cached experiments")
        cached_experiments = experiment.fetch_cached(candidate_data)
        logger.info("Found {} experiments.".format(len(cached_experiments)))
        if len(cached_experiments) > 0:
            summary, seed_data = analyzer.analyze(cached_experiments, seed_data)
            # Remove cached experiments from candidate_data
            candidate_space = candidate_data.index.difference(
                cached_experiments.index, sort=False
            ).tolist()
            candidate_data = candidate_data.loc[candidate_space]
            logger.info("Cached experiments added to seed.")

        # Construct and start loop
        return cls(
            candidate_data=candidate_data, agent=agent, experiment=experiment,
            analyzer=analyzer, seed_data=seed_data,
            heuristic_stopper=5, s3_prefix=s3_prefix,
            logger=logger
        )

    @classmethod
    def from_chemsys_high_quality(cls, chemsys, **kwargs):
        """
        Factory method for generating higher-tier campaigns,
        i.e. with longer walltime, higher quality batch queue,
        and higher number of maximum atoms in candidates

        Args:
            chemsys (str): chemical system for the campaign

        Returns:
            (ProtoDFTCampaign) for campaign corresponding to chemical system

        """
        experiment = OqmdDFTonMC1(timeout=120000, batch_queue="oqmd_prod",
                                  prefix_append="proto-dft-high")
        return cls.from_chemsys(
            chemsys, n_max_atoms=40, experiment=experiment,
            prefix="proto-dft-high/runs", **kwargs)

    def autorun(self):
        """
        Method for running this campaign automatically

        Returns:
            None

        """
        n_max_iter = n_max_iter_heuristics(
            len(self.candidate_data), 10)
        self.auto_loop(
            n_iterations=n_max_iter, monitor=True,
            initialize=True, save_iterations=True
        )


class CloudATFCampaign(Campaign):
    """
    Simple subclass for cloud-based ATF, mostly for testing
    """
    @classmethod
    def from_chemsys(cls, chemsys):
        """
        Args:
            chemsys:
        Returns:
        """
        s3_prefix = "oqmd-atf/runs/{}".format(chemsys)
        df = pd.read_csv(os.path.join(CAMD_TEST_FILES, 'test_df.csv'))
        n_seed = 200  # Starting sample size
        n_query = 10  # This many new candidates are "calculated with DFT" (i.e. requested from Oracle -- DFT)
        agent = RandomAgent(n_query=n_query)
        analyzer = StabilityAnalyzer(hull_distance=0.05)
        experiment = ATFSampler(dataframe=df)
        candidate_data = df
        return cls(candidate_data, agent, experiment, analyzer,
                   create_seed=n_seed, s3_prefix=s3_prefix)

    def autorun(self):
        """
        Runs campaign with standard parameters
        Returns:
            None
        """
        self.auto_loop(initialize=True, n_iterations=3)
        return True


def n_max_iter_heuristics(n_data, n_query, low_bound=5, up_bound=20):
    """
    Helper method to define maximum number of iterations for
    a given campaign.  This is based on the empirical evidence
    in various systems >90% of stable materials are identified
    when 25% of candidates are tested. We also enforce upper
    and lower bounds of 20 and 5 to avoid edge cases with too
    many or too few calculations to run.

    Args:
        n_data (int): number of data points in candidate space
        n_query (int): number of queries allowed in each iteration
        low_bound (int): lower bound allowed for n_max_iter
        up_bound (int): upper bound allowed for n_max_ite

    Returns:
        maximum number of iterations as integer

    """
    _target = round(n_data * 0.25/n_query)
    if _target < low_bound:
        return low_bound
    else:
        return min(_target, up_bound)



#  Copyright (c) 2019 Toyota Research Institute.  All rights reserved.

import unittest
import boto3
import os

os.environ['CAMD_S3_BUCKET'] = 'camd-test'
from taburu.table import ParameterTable
from camd import CAMD_S3_BUCKET
from camd.utils.data import load_default_atf_data, get_oqmd_data_by_chemsys, \
    partition_intercomp
from camd.campaigns.meta_agent import MetaAgentCampaign, StabilityCampaignAnalyzer, \
    META_AGENT_PREFIX
from camd.experiment.agent_simulation import LocalAgentSimulation
from monty.tempfile import ScratchDir
from camd.analysis import StabilityAnalyzer
from camd.agent.base import RandomAgent


def teardown_s3():
    """Tear down test files in s3"""
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(CAMD_S3_BUCKET)
    bucket.objects.filter(Prefix=META_AGENT_PREFIX).delete()


TEST_REGRESSOR_PARAMS = [
    {
        "@class": ["sklearn.linear_model.LinearRegression"],
        "fit_intercept": [True, False],
        "normalize": [True, False]
    },
]


TEST_AGENT_PARAMS = [
    {
        "@class": ["camd.agent.stability.QBCStabilityAgent"],
        "n_query": [10],
        "n_members": list(range(2, 5)),
        "hull_distance": [0.05],
        "training_fraction": [0.4],
        "model": TEST_REGRESSOR_PARAMS
    },
]


RANDOM_TEST_AGENT_PARAMS = [
    {
        "@class": ["camd.agent.base.RandomAgent"],
        "n_query": [2, 3],
    },
]


@unittest.skipUnless(CAMD_S3_BUCKET, "CAMD S3 Bucket not set")
class MetaAgentCampaignTest(unittest.TestCase):
    def tearDown(self):
        teardown_s3()

    def test_initialize_and_update(self):
        agent_pool = ParameterTable(TEST_AGENT_PARAMS)
        dataframe = get_oqmd_data_by_chemsys("Fe-O")
        cand, seed = partition_intercomp(dataframe, n_elements=1)
        analyzer = StabilityAnalyzer()
        experiment = LocalAgentSimulation(
            cand, iterations=5,
            analyzer=analyzer, seed_data=seed
        )

        MetaAgentCampaign.reserve(
            name="test_meta_agent", experiment=experiment,
            agent_pool=agent_pool, analyzer=analyzer
        )
        self.assertRaises(ValueError, MetaAgentCampaign.reserve,
                          "test_meta_agent", dataframe, agent_pool, None)

        agent_pool, data, analyzer = MetaAgentCampaign.load_pickled_objects(
            "test_meta_agent"
        )
        self.assertEqual(len(agent_pool), 12)

        MetaAgentCampaign.update_agent_pool(
            "test_meta_agent",
            TEST_AGENT_PARAMS
        )
        agent_pool, _, _ = MetaAgentCampaign.load_pickled_objects(
            "test_meta_agent"
        )
        self.assertEqual(len(agent_pool), 12)

    def test_run(self):
        agent_pool = ParameterTable(RANDOM_TEST_AGENT_PARAMS)
        # Construct experiment
        dataframe = get_oqmd_data_by_chemsys("Fe-O")
        cand, seed = partition_intercomp(dataframe, n_elements=1)
        experiment = LocalAgentSimulation(
            atf_candidate_data=cand, seed_data=seed,
            analyzer=StabilityAnalyzer(), iterations=10,
        )
        analyzer = StabilityCampaignAnalyzer(checkpoint_indices=[2, 5, 10])
        MetaAgentCampaign.reserve(
            name="test_meta_agent", experiment=experiment,
            agent_pool=agent_pool, analyzer=analyzer
        )
        with ScratchDir('.'):
            print("Testing meta agent")
            campaign = MetaAgentCampaign.from_reserved_name(
                "test_meta_agent", meta_agent=RandomAgent(n_query=1),
            )
            campaign.autorun()
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



# Copyright Toyota Research Institute 2019
"""
Module containing DFT-related experiments, typically
to be run asynchronously with a campaign.
"""

import os
import json
import re
import time
import shlex
import shutil
import subprocess
import traceback
import warnings
from datetime import datetime

import boto3
from botocore.errorfactory import ClientError
from tqdm import tqdm
import pandas as pd
import numpy as np
from monty.os import cd
from monty.tempfile import ScratchDir
from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.core import Composition, Structure
from pymatgen.io.vasp.sets import MPRelaxSet
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.entries.compatibility import MaterialsProjectCompatibility
from camd.experiment.base import Experiment
from camd.utils.data import (
    MP_REFERENCES,
    QMPY_REFERENCES,
    QMPY_REFERENCES_HUBBARD,
    get_chemsys,
    get_common_prefixes,
)

from atomate.vasp.database import VaspCalcDb
from atomate.vasp.config import (
    ADD_WF_METADATA,
    DB_FILE,
    VASP_CMD,
)
from atomate.vasp.powerups import (
    add_common_powerups,
    add_wf_metadata,
)
from atomate.vasp.workflows.base.core import get_wf

from fireworks.core.rocket_launcher import rapidfire


def wf_structure_optimization(structure, wf_name=None, c=None):
    """
    Hacking the atomate wf_structure optimization to allow
    wf_name added to the metadata

    Args:
        structure (Structure)
        wf_name (str)
        c (dict)
    Returns:

    """
    c = c or {}
    vasp_cmd = c.get("VASP_CMD", VASP_CMD)
    db_file = c.get("DB_FILE", DB_FILE)
    user_incar_settings = c.get("USER_INCAR_SETTINGS")

    wf = get_wf(
        structure,
        "optimize_only.yaml",
        vis=MPRelaxSet(
            structure, force_gamma=True, user_incar_settings=user_incar_settings
        ),
        common_params={"vasp_cmd": vasp_cmd, "db_file": db_file},
        wf_metadata={"wf_name": wf_name},
    )

    wf = add_common_powerups(wf, c)

    if c.get("ADD_WF_METADATA", ADD_WF_METADATA):
        wf = add_wf_metadata(wf, structure)

    return wf


class AtomateExperiment(Experiment):
    """
    A class for brokering atomate experiments
    TODO: 1. formation energy solution
    """

    def __init__(
        self,
        launchpad,
        db_file,
        fworker=None,
        atomate_workflow=wf_structure_optimization,
        nlaunches=0,
        max_loops=-1,
        sleep_time=5,
        m_dir=None,
        poll_time=60,
        current_data=None,
        job_status=None,
        history=None,
        launch_from_local=False,
    ):
        """
        Initializes an atomate experiment.

        Args:
            launchpad (LaunchPad): launchpad
            db_file (str): path to atomate db config file
            fworker (FWorker object): fworker
            atomate_workflow (): atomate workflow, default the optimziation wf
            nlaunches (int): 0 means 'until completion', -1 or "infinite" means to loop until max_loops
            max_loops (int): maximum number of loops (default -1 is infinite)
            sleep_time (int): secs to sleep between rapidfire loop iterations
            m_dir (str): the directory in which to loop Rocket running
            poll_time (int): time in seconds to wait in between queries of db
            current_data (pandas.DataFrame): dataframe corresponding to
                currently submitted experiments
            job_status (str): status of the experiment, PENDING or COMPLETED
            history (pandas.DataFrame): history of past experiments
            launch_from_local (bool): whether to launch from local
        """
        self.current_data = current_data
        self.job_status = job_status
        self._history = history or []
        self.launchpad = launchpad
        self.fworker = fworker
        self.nlaunches = nlaunches
        self.max_loops = max_loops
        self.sleep_time = sleep_time
        self.m_dir = m_dir
        self.db_file = db_file
        self.wf = atomate_workflow
        self.poll_time = poll_time
        self.launch_from_local = launch_from_local

    @property
    def db(self):
        """
        Define the db attribute as a property so that the experiment 
        object is picklable. Without this, the SSLContext associated
        with the DB is not picklable.
        """
        return VaspCalcDb.from_db_file(self.db_file).db

    # def save(self, path):
    #     pass

    def update_current_data(self, data):
        """
        Updates current data with dataframe,
        stores old data in history

        Args:
            data (DataFrame):

        Returns:
            None

        """
        if self.current_data is not None:
            current_results = self.get_results()
            self._history.append((self.current_data, current_results))
        self.current_data = data

    def get_results(self):
        """
        Gets the results from the current run.

        Returns:
            (DataFrame): dataframe corresponding to the
                current set of results

        """
        if self.job_status != "COMPLETED":
            self.update_results()
        if self.job_status != "COMPLETED":
            warnings.warn("Some calculations have not finished")
        return self.current_data

    def monitor(self):
        """
        Method for continuously polling for completion
        Returns:
            (str): calculation status string
        """
        while self.job_status != "COMPLETED":
            time.sleep(self.poll_time)
            self.update_results()
            self.print_status()
        return self.job_status

    def update_results(self):
        """
        Update the current data with the latest wf, fw, launch, task information
        """
        task_id = None
        launch_id = None
        task_status = None
        output = None
        input = None
        task_dir = None
        calcs_reversed = None
        delta_e = None
        for structure_id, row in self.current_data.iterrows():
            wf_entry = self.db.workflows.find_one(
                {
                    "metadata.wf_name": row["wf_name"],
                    "fw_states.{}".format(row["fw_id"]): {"$exists": True},
                }
            )
            fw_entry = self.db.fireworks.find_one({"fw_id": row["fw_id"]})
            if len(fw_entry["launches"]) > 0:
                launch_id = fw_entry["launches"][-1]
                launch_entry = self.db.launches.find_one({"launch_id": launch_id})
                if launch_entry["action"]:
                    if "task_id" in launch_entry["action"]["stored_data"]:
                        task_id = launch_entry["action"]["stored_data"]["task_id"]
                        task_entry = self.db.tasks.find_one({"task_id": task_id})
                        task_status = task_entry["state"]
                        output = task_entry["output"]
                        input = task_entry["input"]
                        task_dir = task_entry["dir_name"]
                        calcs_reversed = task_entry["calcs_reversed"]
                        if task_status == "successful":
                            delta_e = self.get_delta_e(task_entry)
            update_data = {
                "task_id": task_id if task_id else None,
                "launch_id": launch_id if launch_id else None,
                "wf_status": wf_entry["state"],
                "task_status": task_status if task_status else None,
                "output": output if output else None,
                "input": input if input else None,
                "task_dir": task_dir if task_dir else None,
                "calcs_reversed": json.dumps(calcs_reversed)
                if calcs_reversed
                else None,
                "final_structure": output["structure"] if output else None,
                "final_energy_per_atom": output["energy_per_atom"] if output else None,
                "delta_e": delta_e,
            }
            update_dataframe_row(self.current_data, structure_id, update_data)
        self._update_job_status()

    def _update_job_status(self):
        """
        Update the job_status flag by checking all wf status
        job_status is "COMPLETED" only when all wf_status are COMPLETED/FIZZLED
        """
        wf_status = self.current_data["wf_status"]
        if np.all([i in ["COMPLETED", "FIZZLED"] for i in wf_status]):
            self.job_status = "COMPLETED"

    def print_status(self):
        """
        Prints current status of experiment according to
        the data in the current_data attribute
        """
        status_string = ""
        for structure_id, row in self.current_data.iterrows():
            status_string += "{}: {}\n".format(structure_id, row["wf_status"])
        print("Calc status:\n{}".format(status_string))
        # print("Timeout is set as {}.".format(self.timeout))

    def submit(self, data):
        """
        Args:
            data (DataFrame): dataframe containing all necessary
                data to conduct the experiment(s).  May be one
                row, may be multiple rows

        Returns:
            None
        """
        self.update_current_data(data)
        new_columns = [
            "wf_spec",
            "wf_name",
            "fw_id",
            "task_id",
            "launch_id",
            "wf_status",  # status of the workflow (ready, waiting, running, fizzeld, completed)
            "task_status",  # status of the task (successful, failed)
            "output",
            "input",
            "task_dir",
            "calcs_reversed",
            "final_structure",
            "final_energy_per_atom",
            "delta_e",  # formation energy per atom
        ]
        for new_column in new_columns:
            self.current_data[new_column] = None
        self.add_wfs()
        if self.launch_from_local:
            self.launch()
        self.job_status = "PENDING"
        return self.job_status

    def add_wfs(self):
        """
        Helper function to add workflows to the launchpad
        and update the wf_name, fw_ids
        """
        for structure_id, row in self.current_data.iterrows():
            wf_name = f"opt_{structure_id}"
            wf = wf_structure_optimization(row["structure"], wf_name=wf_name)
            fw_ids = self.launchpad.add_wf(wf)
            launch_info = {
                "wf_spec": wf,
                "wf_name": wf_name,
                "fw_id": sorted(list(fw_ids.values()))[-1],
            }
            update_dataframe_row(self.current_data, structure_id, launch_info)

    def launch(self):
        """
        Helper method for launching firework
        """
        rapidfire(
            self.launchpad,
            self.fworker,
            nlaunches=self.nlaunches,
            max_loops=self.max_loops,
            sleep_time=self.sleep_time,
            m_dir=self.m_dir,
        )

    @property
    def agg_history(self):
        """
        Aggregated history, i.e. in two single dataframes
        corresponding to "current data" attributes and
        results

        Returns:
            (DataFrame): history of current data
            (DataFrame): history of results

        """
        cd_list, cr_list = zip(*self._history)
        return pd.concat(cd_list), pd.concat(cr_list)

    def get_delta_e(self, task_entry):
        """
        Helper function to get the formation energy from task entry
        Args:
            task_entry (dict): tasks entry
        Returns:
            formation energy (float)
        """
        total_e = task_entry["output"]["energy"]
        composition = Structure.from_dict(task_entry["output"]["structure"]).composition
        psp = task_entry["input"]["pseudo_potential"]
        functional = psp["functional"].capitalize()
        labels = psp["labels"]
        potcar_symbols = [" ".join([functional, i]) for i in labels]
        hubbards = task_entry["input"]["hubbards"]
        return get_mp_formation_energy(
            total_e, composition.__str__(), potcar_symbols, hubbards
        )


class OqmdDFTonMC1(Experiment):
    """
    An experiment class to manage Density Functional theory
    experiments on the MC1, an AWS-batch-based DFT-calculation
    system.
    """

    def __init__(
        self,
        poll_time=60,
        timeout=7200,
        current_data=None,
        job_status=None,
        cleanup_directories=True,
        container_version="oqmdvasp/3",
        batch_queue="oqmd_test_queue",
        use_cached=False,
        prefix_append=None,
    ):
        """
        Initializes an OqmdDFTonMC1 instance

        Args:
            poll_time (int): time in seconds to wait in between queries of aws batch
            timeout (int): time in seconds to wait before killing batch jobs
            current_data (pandas.DataFrame): dataframe corrsponding to current data
            job_status (str): job status
            cleanup_directories (bool): flag to enable cleaning up of DFT results after
                runs are over
            container_version (str): container for mc1, e.g. "oqmdvasp/3" or "gpaw/1" which
                dictates where things will be run
            batch_queue (str): name of aws batch queue to submit to
            use_cached (bool): whether to used cached results from prior campaigns
            prefix_append (str): appended prefix to use for caching and data organization,
                e.g. proto-dft-1
        """
        self.poll_time = poll_time
        self.timeout = timeout
        self.cleanup_directories = cleanup_directories
        self.use_cached = use_cached
        self.batch_queue = batch_queue

        # Build parent directory using TRI_PATH and container/version
        tri_path = os.environ.get("TRI_PATH")
        tri_bucket = os.environ.get("TRI_BUCKET")
        if not (tri_path and tri_bucket):
            raise ValueError(
                "TRI_PATH and TRI_BUCKET must be specified as env "
                "variable to use camd MC1 interface"
            )

        container, version = container_version.split("/")
        parent_dir = os.path.join(
            tri_path,
            "model",
            container,
            version,
            "u",
            "camd",
        )
        if prefix_append is not None:
            if "_" in prefix_append:
                raise ValueError(
                    "Prefix cannot contain underscores for mc1 compatibility"
                )
            parent_dir = os.path.join(parent_dir, prefix_append)
        self.parent_dir = parent_dir

        super().__init__(current_data=current_data, job_status=job_status)

    def _update_job_status(self):
        """
        Updates the aggregate job status according to the
        entries of the current data's "status" column

        Returns:
            None

        """
        all_job_statuses = self.current_data["status"]
        all_jobs_complete = all(
            [status in ["SUCCEEDED", "FAILED"] for status in all_job_statuses]
        )
        if all_jobs_complete:
            self.job_status = "COMPLETED"

    def update_results(self):
        """
        Updates the results by checking the AWS Batch system
        via check_dft_calcs

        Returns:
            None

        """
        # Update data
        self.check_dft_calcs()

        # Update time
        elapsed_time = datetime.utcnow() - self.current_data["start_time"]
        elapsed_time = [e.total_seconds() for e in elapsed_time]
        self.current_data["elapsed_time"] = elapsed_time

        # Update job status
        self._update_job_status()

    def get_results(self):
        """
        Gets the results from the current run.

        Returns:
            (DataFrame): dataframe corresponding to the
                current set of results

        """
        if self.job_status != "COMPLETED":
            self.update_results()
        if self.job_status != "COMPLETED":
            warnings.warn("Some calculations have not finished")
        return self.current_data

    def submit(self, data):
        """
        Args:
            data (DataFrame): dataframe representing structure
                inputs to experiment.  This dataframe must have
                a 'structure' column

        Returns:
            (str): string corresponding to job status

        """
        if self.current_data is not None and self.cleanup_directories:
            self.clean_current_paths()
        self.update_current_data(data)
        # Populate new columns
        new_columns = [
            "path",
            "status",
            "start_time",
            "jobId",
            "jobName",
            "jobArn",
            "result",
            "error",
            "delta_e",
            "elapsed_time",
        ]
        for new_column in new_columns:
            self.current_data[new_column] = None

        self.submit_dft_calcs_to_mc1()
        self.job_status = "PENDING"
        return self.job_status

    def clean_current_paths(self):
        """
        Helper method to clean simulation results in current paths

        Returns:
            None

        """
        for idx, row in self.current_data.iterrows():
            # Clean simulation directories
            sim_path = row["path"].replace("model", "simulation")
            if os.path.exists(sim_path):
                shutil.rmtree(sim_path)

    def print_status(self):
        """
        Prints current status of experiment according to
        the data in the current_data attribute

        Returns:
            None

        """
        status_string = ""
        for structure_id, row in self.current_data.iterrows():
            status_string += "{}: {} {}\n".format(
                structure_id, row["status"], row["elapsed_time"]
            )
        print("Calc status:\n{}".format(status_string))
        print("Timeout is set as {}.".format(self.timeout))

    def monitor(self):
        """
        Method for continuously polling for completion


        Returns:
            (str): calculation status string

        """
        while self.job_status != "COMPLETED":
            time.sleep(self.poll_time)
            self.update_results()
            self.print_status()
            self.kill_lapsed_jobs()

        return self.job_status

    def fetch_cached(self, candidate_data):
        """
        Fetches cached data based on candidate data

        Args:
            candidate_data (pd.DataFrame): Pandas dataframe

        Returns:
            (pandas.DataFrame) dataframe with data filled out

        """
        tri_path = os.environ.get("TRI_PATH")
        tri_bucket = os.environ.get("TRI_BUCKET")
        s3_client = boto3.client("s3")
        # Scrub tri_path and replace model with simulation
        # to get s3 key
        s3_parent = self.parent_dir.replace("model", "simulation")
        s3_parent = s3_parent.replace(tri_path + "/", "")
        cached_experiments = pd.DataFrame()
        # Get all experiment folders
        chemsyses = set([get_chemsys(s) for s in candidate_data["structure"]])
        experiment_dirs = []
        for chemsys in chemsyses:
            chemsys_dirs = get_common_prefixes(
                tri_bucket, os.path.join(s3_parent, chemsys)
            )
            experiment_dirs.extend(chemsys_dirs)
        for structure_id, row in tqdm(
            candidate_data.iterrows(), total=len(candidate_data)
        ):
            if not structure_id.replace("-", "") in experiment_dirs:
                continue
            calc_path = os.path.join(
                s3_parent,
                get_chemsys(row["structure"]),
                structure_id.replace("-", ""),
                "_1/",
            )
            with ScratchDir("."):
                # Figure out whether prior submission exists
                cached_experiments = cached_experiments.append(row)
                # TODO: figure out whether file exists in s3
                # TODO: this is a little crude, could use boto3
                try:
                    # import pdb; pdb.set_trace()
                    vr_path = os.path.join(calc_path, "static", "vasprun.xml")
                    cmd = "aws s3 cp s3://{}/{} .".format(tri_bucket, vr_path)
                    subprocess.call(shlex.split(cmd))
                    vr = Vasprun("vasprun.xml")
                    vr_dict = vr.as_dict()
                    delta_e = get_qmpy_formation_energy(
                        vr_dict["output"]["final_energy_per_atom"],
                        vr_dict["pretty_formula"],
                        1,
                    )
                    data = {
                        "status": "SUCCEEDED",
                        "error": None,
                        "result": vr,
                        "delta_e": delta_e,
                    }
                except Exception as e:
                    error_doc = {}
                    try:
                        err_obj = s3_client.get_object(
                            Bucket=tri_bucket, Key=os.path.join(calc_path, "err")
                        )
                        errtxt = err_obj["Body"].read().decode("utf-8")
                        error_doc.update({"mc1_stderr": errtxt})
                    except ClientError:
                        print("No error file for {}".format(calc_path))
                    error_doc.update(
                        {
                            "camd_exception": "{}".format(e),
                            "camd_traceback": traceback.format_exc(),
                        }
                    )
                    # Dump error docs to avoid Pandas issues with dict values
                    data = {"status": "FAILED", "error": json.dumps(error_doc)}
                update_dataframe_row(
                    cached_experiments, structure_id, data, add_columns=True
                )
        return cached_experiments

    def submit_dft_calcs_to_mc1(self):
        """
        Helper method for submitting DFT calculations to MC1
        """

        # Create run directory
        # uuid_string = str(uuid.uuid4()).replace("-", "")

        if any(["_" in value for value in self.current_data.index]):
            raise ValueError(
                "Structure keys cannot contain underscores for " "mc1 compatibility"
            )

        for structure_id, row in self.current_data.iterrows():
            # Replace structure id in path to avoid confusing mc1
            calc_path = os.path.join(
                self.parent_dir,
                get_chemsys(row["structure"]),
                structure_id.replace("-", ""),
                "_1",
            )
            os.makedirs(calc_path, exist_ok=True)
            with cd(calc_path):
                # Write input cif file and python model file
                row["structure"].to(filename="POSCAR")
                with open("model.py", "w") as f:
                    f.write(MODEL_TEMPLATE)

                # Submit to mc1
                # TODO: ensure this is checked for failure to submit
                print("Submitting job: {}".format(structure_id))
                try:
                    response = subprocess.check_output(
                        [
                            "trisub",
                            "-q",
                            self.batch_queue,
                            "-r",
                            "16000",
                            "-c",
                            "16",
                            "-g",
                            "us-east-1",
                        ]
                    )
                    response = response.decode("utf-8")
                    response = re.findall("({.+})", response, re.DOTALL)[0]
                    data = json.loads(response)
                    data.update(
                        {
                            "path": os.getcwd(),
                            "status": "SUBMITTED",
                            "start_time": datetime.utcnow(),
                        }
                    )
                except subprocess.CalledProcessError as e:
                    print(e.output)
                    data = {
                        "path": os.getcwd(),
                        "status": "FAILED",
                        "error": "failed submission {}".format(e.output),
                    }

                update_dataframe_row(self.current_data, structure_id, data)

    def check_dft_calcs(self):
        """
        Helper function to check DFT calculations via polling
        AWS batch.  Updates current data with latest AWS batch
        response.
        """
        for structure_id, calc in self.current_data.iterrows():
            if calc["status"] in ["SUCCEEDED", "FAILED"]:
                continue
            path = calc["path"]
            print("Checking status of {}: {}".format(path, structure_id))
            aws_cmd = "aws batch describe-jobs --jobs " "--region=us-east-1 {}".format(
                calc["jobId"]
            )
            result = subprocess.check_output(shlex.split(aws_cmd))
            result = json.loads(result)
            aws_status = result["jobs"][0]["status"]
            if aws_status == "SUCCEEDED":
                os.chdir(path)
                subprocess.call("trisync")
                os.chdir("simulation")
                try:
                    vr = Vasprun("static/vasprun.xml")
                    vr_dict = vr.as_dict()
                    delta_e = get_qmpy_formation_energy(
                        vr_dict["output"]["final_energy_per_atom"],
                        vr_dict["pretty_formula"],
                        1,
                    )
                    data = {
                        "status": "SUCCEEDED",
                        "error": None,
                        "result": vr,
                        "delta_e": delta_e,
                    }
                except Exception as e:
                    error_doc = {}
                    if os.path.isfile("err"):
                        with open("err") as errfile:
                            error_doc.update({"mc1_stderr": errfile.read()})
                    error_doc.update(
                        {
                            "camd_exception": "{}".format(e),
                            "camd_traceback": traceback.format_exc(),
                        }
                    )

                    # Dump error docs to avoid Pandas issues with dict values
                    data = {"status": "FAILED", "error": json.dumps(error_doc)}
                os.chdir(path)

            elif aws_status == "FAILED":
                error_doc = {
                    "aws_fail": result["jobs"][0]["attempts"][-1]["statusReason"]
                }
                data = {"status": "FAILED", "error": json.dumps(error_doc)}
            else:
                data = {"status": aws_status}
            update_dataframe_row(self.current_data, structure_id, data)

    def kill_lapsed_jobs(self):
        """
        Method for killing lapsed jobs according to the amount
        of elapsed time and the object's timeout attribute

        Returns:
            None
        """
        running_jobs = self.current_data[self.current_data["status"] == "RUNNING"]
        lapsed_jobs = running_jobs[running_jobs["elapsed_time"] > self.timeout]

        # Kill AWS job
        for structure_id, row in lapsed_jobs.iterrows():
            kill_cmd = (
                "aws batch terminate-job --region=us-east-1 "
                "--job-id {} --reason camd_timeout".format(row["jobId"])
            )
            kill_result = subprocess.check_output(shlex.split(kill_cmd))
            print("{} job killed: ".format(kill_result))
            self.current_data.loc[structure_id, "status"] = "FAILED"
            self.current_data.loc[structure_id, "error"] = "timeout"
        self._update_job_status()


MODEL_TEMPLATE = """
import os

import qmpy
from qmpy.materials.structure import Structure
from qmpy.analysis.vasp.calculation import Calculation
from qmpy import io
import time


# TODO: definitely move this somewhere else, as it's not
#       meant to be imported
def run_oqmd_calculation(poscar_filename):
    starting_structure = io.poscar.read(poscar_filename)

    # Relaxation
    os.mkdir("relax")
    os.chdir("relax")
    calc = Calculation()
    calc.setup(starting_structure, "relaxation")
    os.system("mpirun -n 16 vasp_std")
    # Just in case the mysql server process dies
    os.system("sudo -u mysql mysqld &")
    relaxed_structure = io.poscar.read("CONTCAR")
    os.chdir('..')

    # Relaxation
    os.mkdir("static")
    os.chdir("static")
    calc = Calculation()
    calc.setup(relaxed_structure, "static")
    os.system("mpirun -n 16 vasp_std")
    os.chdir('..')


if __name__ == '__main__':
    run_oqmd_calculation("POSCAR")
"""


def get_qmpy_formation_energy(total_e, formula, n_atoms):
    """
    Helper function to computer qmpy-compatible formation
    energy using reference energies extracted from OQMD

    Args:
        total_e (float): total energy
        formula (str): chemical formula
        n_atoms (int): number of atoms

    Returns:
        (float): qmpy-compatible formation energy

    """
    composition = Composition(formula).fractional_composition
    energy = total_e / n_atoms
    for element, weight in composition.as_dict().items():
        energy -= QMPY_REFERENCES[element] * weight
        if (element in QMPY_REFERENCES_HUBBARD) and ("O" in composition):
            energy -= QMPY_REFERENCES_HUBBARD[element] * weight
    return energy


def get_mp_formation_energy(
    total_e,
    formula,
    potcar_symbols,
    hubbards={},
    explain=False
):
    """
    Helper function to computer mp-compatible formation
    energy using reference energies extracted from MP

    Args:
        total_e (float): total energy (uncorrected)
        formula (str): chemical formula
        potcar_symbols (list): list of potcar symbols
        hubbards (dict): hubbard value, if none, the
                        run_type = 'GGA'
                        else run_type = 'GGA+U'
        explain (bool): whether to print out the explanation
                        of the correction

    Returns:
        (float): mp-compatible formation energy (eV/atom)

    """
    compatibility = MaterialsProjectCompatibility()
    comp = Composition(formula)
    run_type = "GGA+U" if hubbards else "GGA"
    is_hubbard = True if hubbards else False
    entry = ComputedEntry(
        composition=comp,
        energy=total_e,
        parameters={
            "potcar_symbols": potcar_symbols,
            "run_type": run_type,
            "hubbards": hubbards,
            "is_hubbard": is_hubbard,
        },
    )
    # process_entry has a bad try/except block that makes it difficult to catch errors,
    # use this instead
    entry = compatibility.process_entries([entry])[0]
    if explain:
        print(compatibility.explain(entry))
    energy = entry.energy
    for el, occ in comp.items():
        energy -= MP_REFERENCES[el.name] * occ
    return energy / comp.num_atoms


def get_mp_formation_energy_from_m3gnet(total_e, structure, explain=False):
    """
    Hack to get materials-project compatible formation energy from m3gnet
    by overriding potcar/hubbard values with default values from MPRelaxSet
    """
    vaspset = MPRelaxSet(structure)
    potcar_symbols = ["PBE {}".format(sym) for sym in vaspset.potcar_symbols]
    symbols = [s.split()[1] for s in potcar_symbols]
    symbols = [re.split(r"_", s)[0] for s in symbols]
    if not vaspset.incar.get("LDAU", False):
        hubbards = {}
    else:
        us = vaspset.incar.get("LDAUU")
        js = vaspset.incar.get("LDAUJ")
        if len(js) != len(us):
            js = [0] * len(us)
        if len(us) == len(symbols):
            hubbards = {symbols[i]: us[i] - js[i] for i in range(len(symbols))}
        elif sum(us) == 0 and sum(js) == 0:
            hubbards = {}
    return get_mp_formation_energy(
        total_e, structure.composition.formula, potcar_symbols=potcar_symbols,
        hubbards=hubbards, explain=explain)


def update_dataframe_row(dataframe, index, update_dict, add_columns=False):
    """
    Method to update a dataframe row via an update_dictionary
    and an index, similarly to Dict.update()

    Args:
        dataframe (DataFrame): DataFrame for which rows should
            be updated
        index: index value for dataframe
        update_dict ({}): update dictionary for dataframe
        add_columns (bool): whether to add non-existent
            columns to the dataframe

    Returns:
        None

    """
    for key, value in update_dict.items():
        if add_columns and key not in dataframe.columns:
            dataframe[key] = None
        if isinstance(value, dict):
            dataframe.loc[index, key] = json.dumps(value)
        else:
            dataframe.loc[index, key] = value



# Copyright Toyota Research Institute 2019
"""
Module for conducting analysis/postprocessing of
experiments and campaigns.  Contains the Analyzer
object, which performs this function.
"""

import abc
import warnings
import json
import os
import numpy as np
import itertools
import pandas as pd
from camd import tqdm
from qmpy.analysis.thermodynamics.phase import Phase, PhaseData
from qmpy.analysis.thermodynamics.space import PhaseSpace
from multiprocessing import Pool, cpu_count
from pymatgen.core.composition import Composition
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.analysis.phase_diagram import (
    PhaseDiagram,
    PDPlotter,
    tet_coord,
    triangular_coord,
)
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core.structure import Structure
from camd.utils.data import cache_matrio_data, filter_dataframe_by_composition, ELEMENTS
from camd import CAMD_CACHE


class AnalyzerBase(abc.ABC):
    """
    The AnalyzerBase class defines the contract
    for post-processing experiments and reporting
    on campaign state
    """

    def __init__(self):
        """
        Initialize an Analyzer.  Should contain all necessary
        state variables for consistent analysis methods
        """
        self._initial_seed_indices = []

    @abc.abstractmethod
    def analyze(self, campaign):
        """
        Analyze method, performs some operation on new
        experimental results in order to place them
        in the context of the seed data

        Args:
            campaign (Campaign): Campaign to be analyzed

        Returns:
            (DataFrame): dataframe corresponding to the analysis summary
        """


class GenericMaxAnalyzer(AnalyzerBase):
    """
    Generic analyzer that checks new data with a target column against a threshold to be crossed.
    """

    def __init__(self, threshold=0, target="target"):
        """
        Args:
            threshold (int or float): The target values of new acquisitions are compared to find if they are above this
            threshold, to keep track of the performance in sequential framework.
        """
        self.threshold = threshold
        self.target = target
        self.score = []
        self.best_examples = []
        super(GenericMaxAnalyzer, self).__init__()

    def analyze(self, campaign, finalize=False):
        """
        Analyzes the results of an experiment by finding
        the best examples and their scores

        Args:

        Returns:
            (pandas.DataFrame): one-row dataframe summarizing past results
            (pandas.DataFrame): new seed data to be passed to agent

        """
        new_seed = campaign.seed_data.append(campaign.experiment.get_results())
        self.score.append(np.sum(new_seed[self.target] > self.threshold))
        self.best_examples.append(new_seed.loc[new_seed[self.target].idxmax()])
        new_discovery = (
            [self.score[-1] - self.score[-2]]
            if len(self.score) > 1
            else [self.score[-1]]
        )
        summary = pd.DataFrame(
            {
                "score": [self.score[-1]],
                "best_example": [self.best_examples[-1]],
                "new_discovery": new_discovery,
            }
        )
        return summary


# TODO: this isn't yet a true analyzer - still trying to figure out
#   how to implement this consistently
class AnalyzeStructures(AnalyzerBase):
    """
    This class tests if a list of structures are unique. Typically
    used for comparing hypothetical structures (post-DFT relaxation)
    and those from ICSD.
    """

    def __init__(self, structures=None, hull_distance=None):
        """
        Analyzer for structural analysis of jobs

        Args:
            structures ([Structure]): list of a-priori structures to
                compare against
            hull_distance ([float]): hull_distance by which to filter
                results

        """
        self.structures = structures if structures else []
        self.structure_ids = None
        self.unique_structures = None
        self.groups = None
        self.energies = None
        self.against_icsd = False
        self.structure_is_unique = None
        self.hull_distance = hull_distance
        super(AnalyzeStructures, self).__init__()

    def _analyze_structures(
        self, structures=None, structure_ids=None, against_icsd=False, energies=None
    ):
        """
        One encounter of a given structure will be labeled as True, its
        remaining matching structures as False.

        Args:
            structures (list): a list of structures to be compared.
            structure_ids (list): uids of structures, optional.
            against_icsd (bool): whether a comparison to icsd is also made.
            energies (list): list of energies (per atom) corresponding
                to structures. If given, the lowest energy instance of a
                given structure will be return as the unique one. Otherwise,
                there is no such guarantee. (optional)

        Returns:
            ([bool]) list of bools corresponding to the given list of
                structures corresponding to uniqueness
        """
        self.structures = structures
        self.structure_ids = structure_ids
        self.against_icsd = against_icsd
        self.energies = energies

        smatch = StructureMatcher()
        self.groups = smatch.group_structures(structures)
        self.structure_is_unique = []

        if self.energies:
            for i in range(len(self.groups)):
                self.groups[i] = [
                    x
                    for _, x in sorted(
                        zip(
                            [
                                self.energies[self.structures.index(s)]
                                for s in self.groups[i]
                            ],
                            self.groups[i],
                        )
                    )
                ]

        self._unique_structures = [i[0] for i in self.groups]
        for s in structures:
            if s in self._unique_structures:
                self.structure_is_unique.append(True)
            else:
                self.structure_is_unique.append(False)
        self._not_duplicate = self.structure_is_unique

        if self.against_icsd:
            structure_file = "oqmd1.2_exp_based_entries_structures.json"
            cache_matrio_data(structure_file)
            with open(os.path.join(CAMD_CACHE, structure_file), "r") as f:
                icsd_structures = json.load(f)
            chemsys = set()
            for s in self._unique_structures:
                chemsys = chemsys.union(set(s.composition.as_dict().keys()))

            self.icsd_structs_inchemsys = []
            for k, v in icsd_structures.items():
                try:
                    s = Structure.from_dict(v)
                    elems = set(s.composition.as_dict().keys())
                    if elems == chemsys:
                        self.icsd_structs_inchemsys.append(s)
                # TODO: can we make this exception more specific,
                #  do we have an example where this fails?
                except Exception as e:
                    warnings.warn("Unable to process structure {}".format(k))
                    warnings.warn("Error: {}".format(e))

            self.matching_icsd_strs = []
            for i in range(len(structures)):
                if self.structure_is_unique[i]:
                    match = None
                    for s2 in self.icsd_structs_inchemsys:
                        if smatch.fit(self.structures[i], s2):
                            match = s2
                            break
                    self.matching_icsd_strs.append(
                        match
                    )  # store the matching ICSD structures.
                else:
                    self.matching_icsd_strs.append(None)

            # Flip matching bools, and create a filter
            self._icsd_filter = [not i for i in self.matching_icsd_strs]
            self.structure_is_unique = (
                np.array(self.structure_is_unique) * np.array(self._icsd_filter)
            ).tolist()
            self.unique_structures = list(
                itertools.compress(self.structures, self.structure_is_unique)
            )
        else:
            self.unique_structures = self._unique_structures

        # We store the final list of unique structures as unique_structures.
        # We return a corresponding list of bool to the initial structure
        # list provided.
        return self.structure_is_unique

    def analyze(self, campaign):
        """
        Analysis method for structures, not yet implemented

        Args:
            campaign (Campaign): campaign object to be analyzed

        Returns:
            (pd.DataFrame): summary of results from structure analysis

        """
        raise NotImplementedError

    def analyze_vaspqmpy_jobs(self, jobs, against_icsd=False, use_energies=False):
        """
        Useful for analysis integrated as part of a campaign itself

        Args:
            jobs (pd.DataFrame): dataframe of DFT experiment results
            against_icsd (bool): whether to validate against ICSD or not

        Returns:
        """
        self.structure_ids = []
        self.structures = []
        self.energies = []
        for j, r in jobs.iterrows():
            if r["status"] == "SUCCEEDED":
                # This is a switch for OQMD vs. MP
                if "output" in r:
                    final_structure = r["output"]["structure"]
                    self.structures.append(final_structure)
                    self.energies.append(r["output"]["energy_per_atom"])
                    self.structure_ids.append(j)
                else:
                    final_structure = r["result"].final_structure
                    self.structures.append(final_structure)
                    self.structure_ids.append(j)
                    self.energies.append(
                        r["result"].final_energy / len(final_structure)
                    )
        if use_energies:
            return self._analyze_structures(
                self.structures, self.structure_ids, against_icsd, self.energies
            )
        else:
            return self._analyze_structures(
                self.structures, self.structure_ids, against_icsd
            )


class StabilityAnalyzer(AnalyzerBase):
    """
    Analyzer object for stability campaigns
    """

    def __init__(
        self,
        hull_distance=0.05,
        parallel=cpu_count(),
        entire_space=False,
        plot=True,
        initial_seed_indices=None,
    ):
        """
        The Stability Analyzer is intended to analyze DFT-result
        data in the context of a global compositional seed in
        order to determine phase stability.

        Args:
            hull_distance (float): distance above hull below
                which to deem a material "stable"
            parallel (bool): flag for whether or not
                multiprocessing is to be used
            # TODO: is there ever a case where you would
            #       would want to do the entire space?
            entire_space (bool): flag for whether to analyze
                entire space of results or just new chemical
                space
            plot (bool): whether to generate plot as part of
                standard analyze sequence
        """
        self.hull_distance = hull_distance
        self.parallel = parallel
        self.entire_space = entire_space
        self.space = None
        self.plot = plot
        self.initial_seed_indices = initial_seed_indices
        self.current_analysis = None
        super(StabilityAnalyzer, self).__init__()

    @staticmethod
    def get_phase_space(dataframe):
        """
        Gets PhaseSpace object associated with dataframe

        Args:
            dataframe (DataFrame): dataframe with columns "Composition"
                containing formula and "delta_e" containing
                formation energy per atom
        """
        phases = []
        for data in dataframe.iterrows():
            phases.append(
                Phase(
                    data[1]["Composition"],
                    energy=data[1]["delta_e"],
                    per_atom=True,
                    description=data[0],
                )
            )
        for el in ELEMENTS:
            phases.append(Phase(el, 0.0, per_atom=True))

        pd = PhaseData()
        pd.add_phases(phases)
        space = PhaseSpaceAL(bounds=ELEMENTS, data=pd)
        return space

    @staticmethod
    def add_stability(df, hull_distance=0.2, parallel=cpu_count()):
        """
        Adds stability to a dataframe

        Args:
            df (pd.DataFrame): dataframe of entries with `delta_e` field
            hull_distance (float): distance to the hull to classify as stable
            parallel (int): number of parallel processes to use

        Returns:
            (pd.DataFrame): dataframe of stabilities and is_stable

        """
        space = StabilityAnalyzer.get_phase_space(df)
        new_phases = [p for p in space.phases if p.description in df.index]

        space.compute_stabilities(phases=new_phases, ncpus=parallel)

        # Compute new stabilities and update new seed, note that pandas will complain
        # if the index is not explicit due to multiple types (e. g. ints for OQMD
        # and strs for prototypes)
        new_data = pd.DataFrame(
            {"stability": [phase.stability for phase in new_phases]},
            index=[phase.description for phase in new_phases],
        )
        new_data["is_stable"] = new_data["stability"] <= hull_distance
        return new_data

    def analyze(self, campaign, finalize=False):
        """
        Args:
            campaign (Campaign): CAMD campaign
            finalize (bool): whether or not analysis is final

        Returns:
            (DataFrame): summary of the process, i. e. of
                the increment or experimental results

        """
        # On first run get initial seed indices
        if self.initial_seed_indices is None:
            self.initial_seed_indices = campaign.seed_data.index
        # Check for new results
        new_experimental_results = campaign.experiment.get_results()
        new_comp = new_experimental_results["Composition"].sum()
        new_experimental_results = new_experimental_results.dropna(subset=["delta_e"])
        if not finalize:
            new_seed = campaign.seed_data.append(new_experimental_results)
        else:
            new_seed = campaign.seed_data

        # Aggregate seed_data and new experimental results
        include_columns = ["Composition", "delta_e"]
        filtered = new_seed[include_columns].drop_duplicates(keep="last").dropna()

        if not self.entire_space:
            # Constrains the phase space to that of the target compounds.
            # More efficient when searching in a specified chemistry,
            # less efficient if larger spaces are without specified chemistry.
            filtered = filter_dataframe_by_composition(filtered, new_comp)

        new_data = self.add_stability(
            filtered, hull_distance=self.hull_distance, parallel=self.parallel
        )

        # # TODO: This is implicitly adding "stability", and "is_stable" columns
        #       but could be handled more gracefully
        if "stability" not in new_seed.columns:
            new_seed = pd.concat([new_seed, new_data], axis=1, sort=False)
        else:
            new_seed.update(new_data)

        # Write hull figure to disk
        if self.plot:
            self.plot_hull(
                filtered, new_experimental_results.index, filename="hull.png"
            )
        self.current_analysis = new_data
        # Compute summary metrics
        summary = self.get_summary(
            new_seed,
            new_experimental_results.index,
            initial_seed_indices=self.initial_seed_indices,
        )
        return summary

    @staticmethod
    def get_summary(new_seed, new_ids, initial_seed_indices=None):
        """
        Gets summary row for given experimental results after
        preliminary stability analysis.  This is not meant
        to provide the basis for a generic summary method
        and is particular to the StabilityAnalyzer.

        Args:
            new_seed (DataFrame): dataframe corresponding to
                new processed seed
            new_ids ([]): list of index values for those
                experiments that are "new"
            initial_seed_indices ([]): indices of the initial
                seed

        Returns:
            (DataFrame): dataframe summarizing processed
                experimental data including values for
                how many materials were discovered

        """
        # TODO: Right now analyzers don't know anything about the history
        #       of experiments, so can be difficult to determine marginal
        #       value of a given experimental run
        processed_new = new_seed.loc[new_ids]
        initial_seed_indices = (
            initial_seed_indices if initial_seed_indices is not None else []
        )
        total_discovery = new_seed.loc[
            ~new_seed.index.isin(initial_seed_indices)
        ].is_stable.sum()
        return pd.DataFrame(
            {
                "new_candidates": [len(processed_new)],
                "new_discovery": [processed_new.is_stable.sum()],
                "total_discovery": [total_discovery],
            }
        )

    def plot_hull(self, df, new_result_ids, filename=None, finalize=False):
        """
        Generate plots of convex hulls for each of the runs

        Args:
            df (DataFrame): dataframe with formation energies and formulas
            new_result_ids ([]): list of new result ids (i. e. indexes
                in the updated dataframe)
            filename (str): filename to output, if None, no file output
                is produced
            finalize (bool): flag indicating whether to include all new results

        Returns:
            (pyplot): plotter instance
        """
        # Generate all entries
        total_comp = Composition(df["Composition"].sum())
        if len(total_comp) > 4:
            warnings.warn("Number of elements too high for phase diagram plotting")
            return None
        filtered = filter_dataframe_by_composition(df, total_comp)
        filtered = filtered[["delta_e", "Composition"]]
        filtered = filtered.dropna()

        # Create computed entry column with un-normalized energies
        filtered["entry"] = [
            ComputedEntry(
                Composition(row["Composition"]),
                row["delta_e"] * Composition(row["Composition"]).num_atoms,
                entry_id=index,
            )
            for index, row in filtered.iterrows()
        ]

        ids_prior_to_run = list(set(filtered.index) - set(new_result_ids))
        if not ids_prior_to_run:
            warnings.warn("No prior data, prior phase diagram cannot be constructed")
            return None

        # Create phase diagram based on everything prior to current run
        entries = filtered.loc[ids_prior_to_run]["entry"].dropna()

        # Filter for nans by checking if it's a computed entry
        pg_elements = sorted(total_comp.keys())
        pd = PhaseDiagram(entries, elements=pg_elements)
        plotkwargs = {
            "markerfacecolor": "white",
            "markersize": 7,
            "linewidth": 2,
        }
        if finalize:
            plotkwargs.update({"linestyle": "--"})
        else:
            plotkwargs.update({"linestyle": "-"})
        plotter = PDPlotter(pd, backend="matplotlib", **plotkwargs)

        getplotkwargs = {"label_stable": False} if finalize else {}
        plot = plotter.get_plot(**getplotkwargs)

        # Get valid results
        valid_results = [
            new_result_id
            for new_result_id in new_result_ids
            if new_result_id in filtered.index
        ]

        if finalize:
            # If finalize, we'll reset pd to all entries at this point to
            # measure stabilities wrt. the ultimate hull.
            pd = PhaseDiagram(filtered["entry"].values, elements=pg_elements)
            plotter = PDPlotter(
                pd,
                backend="matplotlib",
                **{"markersize": 0, "linestyle": "-", "linewidth": 2}
            )
            plot = plotter.get_plot(plt=plot)

        for entry in filtered["entry"][valid_results]:
            decomp, e_hull = pd.get_decomp_and_e_above_hull(entry, allow_negative=True)
            if e_hull < self.hull_distance:
                color = "g"
                marker = "o"
                markeredgewidth = 1
            else:
                color = "r"
                marker = "x"
                markeredgewidth = 1

            # Get coords
            coords = [entry.composition.get_atomic_fraction(el) for el in pd.elements][
                1:
            ]
            if pd.dim == 2:
                coords = coords + [pd.get_form_energy_per_atom(entry)]
            if pd.dim == 3:
                coords = triangular_coord(coords)
            elif pd.dim == 4:
                coords = tet_coord(coords)
            plot.plot(
                *coords,
                marker=marker,
                markeredgecolor=color,
                markerfacecolor="None",
                markersize=11,
                markeredgewidth=markeredgewidth
            )

        if filename is not None:
            plot.savefig(filename, dpi=70)
        plot.close()

    def finalize(self, campaign):
        """
        Post-processing a dft campaign
        """
        # self.analyze(campaign, finalize=True)
        all_submitted, all_results = campaign.experiment.agg_history
        old_results = campaign.seed_data.drop(all_results.index, errors="ignore")
        new_results = campaign.seed_data.drop(old_results.index)
        st_a = StabilityAnalyzer(
            hull_distance=self.hull_distance,
            parallel=self.parallel,
            entire_space=False,
            plot=False,
        )

        # Having calculated stabilities again, we plot the overall hull.
        # Filter by chemsys
        new_comp = new_results["Composition"].sum()
        filtered = filter_dataframe_by_composition(campaign.seed_data, new_comp)
        filtered = filtered.dropna(subset=["delta_e"])

        st_a.plot_hull(
            filtered,
            all_submitted.index,
            filename="hull_finalized.png",
            finalize=True,
        )

        stabs = self.add_stability(
            filtered, hull_distance=self.hull_distance, parallel=self.parallel
        )
        stable_discovered = stabs[stabs["is_stable"].fillna(False)]

        # Analyze structures if present in experiment
        if "structure" in all_results.columns:
            s_a = AnalyzeStructures()
            s_a.analyze_vaspqmpy_jobs(all_results, against_icsd=True, use_energies=True)
            unique_s_dict = {}
            for i in range(len(s_a.structures)):
                if s_a.structure_is_unique[i] and (
                    s_a.structure_ids[i] in stable_discovered.index
                ):
                    unique_s_dict[s_a.structure_ids[i]] = s_a.structures[i]

            with open("discovered_unique_structures.json", "w") as f:
                json.dump(dict([(k, s.as_dict()) for k, s in unique_s_dict.items()]), f)

            with open("structure_report.log", "w") as f:
                f.write("consumed discovery unique_discovery duplicate in_icsd \n")
                f.write(
                    str(len(all_submitted))
                    + " "
                    + str(len(stable_discovered))
                    + " "
                    + str(len(unique_s_dict))
                    + " "
                    + str(len(s_a.structures) - sum(s_a._not_duplicate))
                    + " "
                    + str(sum([not i for i in s_a._icsd_filter]))
                )
        return True


class PhaseSpaceAL(PhaseSpace):
    """
    Modified qmpy.PhaseSpace for GCLP based stability computations
    TODO: basic multithread or Gurobi for gclp
    """

    def compute_stabilities(self, phases, ncpus=cpu_count()):
        """
        Calculate the stability for every Phase.

        Args:
            phases ([Phase]): list of Phases for which to compute
                stability
            ncpus (int): number of cpus to use, i. e. processes
                to use

        Returns:
            ([float]) stability values for all phases
        """
        self.update_phase_dict(ncpus=ncpus)
        if ncpus > 1:
            with Pool(ncpus) as pool:
                stabilities = pool.map(self.compute_stability, phases)
            # Pool doesn't always modify the phases directly,
            # so assign stability after
            for phase, stability in zip(phases, stabilities):
                phase.stability = stability
        else:
            stabilities = [self.compute_stability(phase) for phase in tqdm(phases)]

        return stabilities

    def compute_stability(self, phase):
        """
        Computes stability for a given phase in the phase
        diagram

        Args:
            phase (Phase): phase for which to compute
                stability

        Returns:
            (float): stability of given phase

        """
        # If the phase name (formula) is in the set of minimal
        # phases by formula, compute it relative to that minimal phase
        if phase.name in self.phase_dict:
            phase.stability = (
                phase.energy
                - self.phase_dict[phase.name].energy
                + self.phase_dict[phase.name].stability
            )
        else:
            phase.stability = self._compute_stability_gclp(phase)

        return phase.stability

    def _compute_stability_gclp(self, phase):
        """
        Computes stability using gclp.  The function
        is still a little unstable, so we use a blank
        try-except to let it do what it can.

        Args:
            phase (Phase): phase for which to compute
                stability using gclp

        Returns:
            (float): stability

        """
        try:
            phase.stability = phase.energy - self.gclp(phase.unit_comp)[0]
        # TODO: do we have an example where this fails?  Can we provide
        #  a more concrete exception?
        except Exception as e:
            print(phase, "stability determination failed, error {}".format(e))
            phase.stability = np.nan
        return phase.stability

    def update_phase_dict(self, ncpus=cpu_count()):
        """
        Function to update the phase dict associated with
        the PhaseSpaceAL

        Args:
            ncpus (int): number of processes to use

        Returns:
            (None)

        """
        uncomputed_phases = [
            phase for phase in self.phase_dict.values() if phase.stability is None
        ]
        if ncpus > 1:
            # Compute stabilities, then update, pool doesn't modify attribute
            with Pool(ncpus) as pool:
                stabilities = pool.map(self._compute_stability_gclp, uncomputed_phases)
            for phase, stability in zip(uncomputed_phases, stabilities):
                phase.stability = stability
        else:
            for phase in uncomputed_phases:
                self._compute_stability_gclp(phase)
        assert (
            len(
                [phase for phase in self.phase_dict.values() if phase.stability is None]
            )
            == 0
        )


class GenericATFAnalyzer(AnalyzerBase):
    """
    Generic analyzer provide AL metrics analysis, including:

    deALM: decision efficiency metric
    anyALM: any discover from top n percentile catalyst
    allALM: fraction of top m percentile
    simALM: similarity between seed_df and new candidate

    This is only used in simulated discovery campaign,
    aggregated selection of candidates from one specific agent
    should be passed in
    """

    def __init__(self, percentile=0.01):
        """
        Notice the default of SL is to maximize a value, if the target value is
        overpotential, please remember to negate the target values

        Args:
            percentile (float):
                top percentile of candidates considered as "good"
        """
        self.percentile = percentile
        super(GenericATFAnalyzer, self).__init__()

    def analyze(self, campaign, finalize=False):
        """
        run analysis one by one

        Args:
            campaign (Campaign): campaign object to analyze
            finalize (bool): whether this analysis is the final one

        Returns:
            (DataFrame): dataframe with summary stats from campaign

        """
        seed_data = campaign.seed_data
        new_experimental_results = campaign.experiment.get_results()
        summary = pd.DataFrame()
        unsampled_candidates = campaign.experiment.dataframe.loc[
            campaign.candidate_data.index
        ]
        all_results = unsampled_candidates.append(
            campaign.experiment.dataframe.loc[campaign.seed_data.index]
        )
        unsampled_candidates = unsampled_candidates.drop(new_experimental_results.index)
        de_alms = self.get_de_alm(new_experimental_results, unsampled_candidates)
        summary["deALM"] = [de_alms]

        any_alm = self.get_any_alm(
            new_experimental_results,
            all_results,
            percentile=self.percentile,
        )
        summary["anyALM"] = [any_alm]

        all_sampled = seed_data.append(new_experimental_results).drop_duplicates()
        all_alm = self.get_all_alm(
            all_sampled,
            all_results,
            percentile=self.percentile,
        )
        summary["allALM"] = [all_alm]
        # simALM_val = self.gather_simALM(
        #     new_experimental_results.copy(), seed_data.copy()
        # )
        # summary["simALM"] = [simALM_val]

        return summary

    @staticmethod
    def get_de_alm(samples, unsampled_results):
        """
        Compute the decision efficiency ALM for one agent.

        de_ALM reflects the percentile of a sample relative
        the remaining unsampled results, and is computed as
        2*f_i - 1 in order to adopt a range from [-1, 1] which
        can be easily compared to the random case (for which
        de_ALM should be 0.)

        If the latest sample was the "best" choice, de_ALM is 1,
        if it was the worst, de_ALM is -1.

        Args:
            samples (pd.DataFrame): latest samples
            unsampled_results (pd.DataFrame): not yet sampled results
                from ATF simulation

        Returns:
            (pd.DataFrame): de_ALM
        """
        # go through each selected candidate, find the rank
        de_ALMs = []
        for idx, sample in samples.iterrows():
            temp = unsampled_results.append(sample)
            sample_rank_pct = temp["target"].rank(pct=True).loc[idx]
            de_ALMs.append(2 * sample_rank_pct - 1)

        return de_ALMs

    @staticmethod
    def get_any_alm(samples, all_results, percentile):
        """
        Compute the any ALM for one agent, which is a boolean
        describing whether or not some sample from the samples thus
        far is present in the top n percentile of materials, e.g.
        if some sample in samples is in the top 1% of all results
        returns True.  Note that the any_alm from Rohr et al. is
        the first iteration in which this is true for a given
        random initialization of an SL campaign.

        Args:
            samples (pd.DataFrame):
                selection of candidiates in the right order
            all_results (pd.DataFrame):
                all_results to use for percentile determination
            percentile (float):
                top percentile of candidates considered as "good"

        Returns:
            (bool): True if any of the samples are in the top
                percentile of all_results, False otherwise
        """
        sample_ranks = all_results["target"].rank(pct=True).loc[samples.index]
        return (sample_ranks > 1 - percentile).any()

    @staticmethod
    def get_all_alm(all_samples, all_possible_samples, percentile):
        """
        Compute the all ALM for one agent, defined as the fraction of
        samples above the target percentile which are contained in
        the samples taken up to this point. For example, if 50% of
        the possible catalysts in the top 1% overpotential have
        been sampled, allALM is 0.5.

        Args:
            all_samples (pd.DataFrame): all taken samples up to this point
            all_possible_samples (pd.DataFrame): all possible samples to
                compute percentiles from
            percentile (float):
                top percentile of candidates considered as "good"
        Returns:
            (float): percentage of top percentile candidates contained
                in aggregated sample set up to this point
        """
        # sort df using target values
        ranked = all_possible_samples["target"].rank(pct=True)
        top_percentile = ranked[ranked > 1 - percentile]
        all_alm = len(top_percentile.index.intersection(all_samples.index)) / len(
            top_percentile
        )
        return all_alm

    # joseph.montoya: I don't know what this is, so I'm going to comment it for now
    #   it looks like it might be a metric related to whether the most recent samples are
    #   inside or outside the domain?
    # def gather_simALM(self, new_experimental_results, seed_data):
    #     """
    #     compute the sim ALM for one agent
    #     Args:
    #         new_experimental_results (pd.DataFrame):
    #             selection of candidiates in the right order
    #         seed_data (pd.DataFrame):
    #             seed data used before first-round CAMD selection starts
    #     Returns:
    #         simALM (np.array):
    #             sim ALM for one specific agent
    #     """
    #     simALM = []
    #     # drop the target values so that only features are left
    #     seed_data.drop("target", inplace=True, axis=1)
    #     new_experimental_results.drop("target", inplace=True, axis=1)
    #     new_experimental_results.reset_index(inplace=True, drop=True)
    #     # go through the new candidate one by one
    #     for i in range(new_experimental_results.shape[0]):
    #         decision_fecture_vector = np.array(new_experimental_results.loc[i].tolist())
    #         seed_feature_vector = np.array(seed_data.values.tolist())
    #         similarity = (
    #             np.linalg.norm(seed_feature_vector - decision_fecture_vector)
    #             / seed_data.shape[0]
    #         )
    #         simALM.append(similarity)
    #         seed_data = pd.concat(
    #             [seed_data, new_experimental_results[i: (i + 1)]], axis=0
    #         )
    #     simALM = np.array(simALM)
    #     return simALM


class MultiAnalyzer(AnalyzerBase):
    """
    The multi-fidelity analyzer.
    """

    def __init__(
        self,
        target_prop,
        prop_range,
        total_expt_queried=0,
        total_expt_discovery=0,
        analyze_cost=False,
        total_cost=0.0,
    ):
        """
        Args:
            target_prop (str)        The name of the target property, e.g. "bandgap".
            prop_range (list)        The range of the target property that is considered
                                     ideal.
            total_expt_queried (int) The total experimental queries after nth iteration.
            tot_expt_discovery (int) The total experimental discovery after nth iteration.
            analyze_cost (bool)      If the input has "cost" column, also analyze that information.
            total_cost(float)        The total cost of the hypotheses after nth iteration.
        """
        self.target_prop = target_prop
        self.prop_range = prop_range
        self.total_expt_queried = total_expt_queried
        self.total_expt_discovery = total_expt_discovery
        self.analyze_cost = analyze_cost
        self.total_cost = total_cost

    def _filter_df_by_prop_range(self, df):
        """
        Helper function that filters df by property range

        Args:
            df   A pd.Dataframe to be filtered.
        """
        return df.loc[
            (df[self.target_prop] >= self.prop_range[0])
            & (df[self.target_prop] <= self.prop_range[1])
        ]

    def analyze(self, campaign, finalize=False):
        """
        Analysis method for campaign

        Args:
            campaign (Campaign):
            finalize (bool):

        Returns:
            (DataFrame): pandas dataframe with summary of the results

        """
        return self._analyze_results(
            campaign.experiment.get_results(), campaign.seed_data
        )

    def _analyze_results(self, new_experimental_results, seed_data):
        """
        Analyze results of multi-fidelity data

        Args:
            new_experimental_results (pd.DataFrame):
            seed_data (pd.DataFrame):

        Returns:
            (pd.DataFrame): summary of last iteration
            (pd.DataFrame): new seed data

        """
        new_expt_hypotheses = new_experimental_results.loc[
            new_experimental_results["expt_data"] == 1
        ]
        new_discoveries = self._filter_df_by_prop_range(new_expt_hypotheses)

        # total discovery = up to (& including) the current iteration
        self.total_expt_queried += len(new_expt_hypotheses)
        self.total_expt_discovery += len(new_discoveries)
        if self.total_expt_queried != 0:
            success_rate = self.total_expt_discovery / self.total_expt_queried
        else:
            success_rate = 0

        summary = pd.DataFrame(
            {
                "expt_queried": [len(new_expt_hypotheses)],
                "total_expt_queried": [self.total_expt_queried],
                "new_discovery": [len(new_discoveries)],
                "total_expt_discovery": [self.total_expt_discovery],
                "total_regret": [self.total_expt_queried - self.total_expt_discovery],
                "success_rate": [success_rate],
            }
        )

        if self.analyze_cost:
            iter_cost = np.sum(new_experimental_results["cost"])
            self.total_cost += iter_cost
            summary["iteration_cost"] = [iter_cost]
            summary["total_cost"] = [self.total_cost]
            if self.total_expt_discovery != 0:
                average_cost_per_discovery = self.total_cost / self.total_expt_discovery
            else:
                average_cost_per_discovery = np.nan
            summary["average_cost_per_discovery"] = [average_cost_per_discovery]
        return summary



import unittest
import uuid
import logging
import json
from datetime import datetime, timedelta
import time

import boto3
from watchtower import CloudWatchLogHandler
from camd.utils.logging import KinesisStreamHandler


# TODO: These tests are only for development purposes
#       and are not being run in CI

class KinesisLoggerTest(unittest.TestCase):
    @unittest.skip
    def test_kinesis_logging(self):
        logger = logging.getLogger("test")
        kinesis_handler = KinesisStreamHandler("camd-events", "test/python-unit-test")
        logger.addHandler(kinesis_handler)
        logger.setLevel("INFO")
        event = {"iteration": 1,
                 "id": str(uuid.uuid4()),
                 "event": "agentGotHypotheses",
                 "metadata": {"chemsys": "Ag-Be"}
                 }
        logger.info(json.dumps(event))

    @unittest.skip
    def test_cloudwatch_logging(self):
        logger = logging.getLogger("test")
        uid = str(uuid.uuid4())
        kinesis_handler = CloudWatchLogHandler(
            log_group="/camd/test/",
            stream_name="python-unittest")
        logger.addHandler(kinesis_handler)
        logger.setLevel("INFO")
        event = {"iteration": 1,
                 "id": uid,
                 "event": "agentGotHypotheses",
                 "metadata": {"chemsys": "Ag-Be"},
                 "timestamp": datetime.now().timestamp()
                 }
        logger.info(json.dumps(event))
        time.sleep(10)

        # Query logs
        client = boto3.client("logs")
        start_query_response = client.get_log_events(
            logGroupName="/camd/test/",
            logStreamName="python-unittest",
            startTime=int((datetime.now() - timedelta(seconds=20)).timestamp())*1000,
            endTime=int(datetime.now().timestamp())*1000,
            # queryString="fields @timestamp, @message | filter streamName python-unittest"
        )


if __name__ == '__main__':
    unittest.main()


import unittest
import pandas
import numpy as np

from camd.agent.generic import RegressorAgent
from camd.agent.base import SequentialAgent


class GenericAgentsTest(unittest.TestCase):
    def setUp(self):
        X = np.linspace(-5, 5, 5)
        Y = 2 * X + 1

        self.seed_data = pandas.DataFrame(
            data=np.vstack((X, Y)).T, index=range(5), columns=["domain", "target"]
        )

        candidate_X = np.linspace(-10, 10, 10)
        self.candidate_data = pandas.DataFrame(data=candidate_X, columns=["domain"])

    def test_linear_agent(self):
        """
        Verify that linear agent can fit a linear
        function and recover the minimizing argument to new data.
        """
        agent = RegressorAgent.from_linear(n_query=10)
        predictions = agent.get_hypotheses(
            seed_data=self.seed_data, candidate_data=self.candidate_data
        )
        self.assertEqual(predictions["domain"][9], 10)

    def test_sequential_agent(self):
        agents = [
            RegressorAgent.from_linear(n_query=10),
            RegressorAgent.from_linear(n_query=5)
        ]
        agent = SequentialAgent(agents=agents)
        hypotheses = agent.get_hypotheses(self.candidate_data, self.seed_data)
        self.assertEqual(len(hypotheses), 5)
